"""use longtext on bulk_actions.body in mysql

Revision ID: 29b539ebc72a
Revises: f495c27d5019
Create Date: 2026-01-06 07:02:58.437749

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import inspect
from sqlalchemy.dialects.mysql import LONGTEXT

# revision identifiers, used by Alembic.
revision = "29b539ebc72a"
down_revision = "f495c27d5019"
branch_labels = None
depends_on = None


def upgrade():
    inspector = inspect(op.get_bind())
    if "mysql" not in inspector.dialect.dialect_description:
        return

    op.alter_column(
        table_name="bulk_actions",
        column_name="body",
        nullable=True,
        type_=sa.types.Text().with_variant(LONGTEXT, "mysql"),
        existing_type=sa.types.Text(),
    )


def downgrade():
    inspector = inspect(op.get_bind())
    if "mysql" not in inspector.dialect.dialect_description:
        return

    op.alter_column(
        table_name="bulk_actions",
        column_name="body",
        nullable=True,
        type_=sa.types.Text(),
        existing_type=sa.types.Text().with_variant(LONGTEXT, "mysql"),
    )



"""As an open source project, we collect usage statistics to inform development priorities.
For more information, check out the docs at https://docs.dagster.io/about/telemetry'.

To see the logs we send, inspect $DAGSTER_HOME/logs/ if $DAGSTER_HOME is set or ~/.dagster/logs/

See class TelemetryEntry for logged fields.

For local development:
  Spin up local telemetry server and set the environment variable DAGSTER_TELEMETRY_URL = 'http://localhost:3000/actions'
  To test RotatingFileHandler, can set MAX_BYTES = 500
"""

import datetime
import hashlib
import inspect
import os
import uuid
from collections.abc import Callable, Mapping, Sequence
from functools import wraps
from typing import TYPE_CHECKING, Any, Optional, TypeVar, overload

from dagster_shared.telemetry import (
    DAGSTER_HOME_FALLBACK,
    TelemetryEntry,
    TelemetrySettings,
    dagster_home_if_set,
    get_or_set_instance_id,
    get_or_set_user_id,
    log_telemetry_action,
    write_telemetry_log_line,
)
from typing_extensions import ParamSpec

import dagster._check as check
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicyType
from dagster._core.definitions.backfill_policy import BackfillPolicyType
from dagster._core.definitions.job_base import IJob
from dagster._core.definitions.reconstruct import (
    ReconstructableJob,
    ReconstructableRepository,
    get_ephemeral_repository_name,
)
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.events import DagsterEvent
from dagster._core.execution.context.system import PlanOrchestrationContext
from dagster._core.execution.plan.objects import StepSuccessData
from dagster._core.instance import DagsterInstance
from dagster._utils.merger import merge_dicts

if TYPE_CHECKING:
    from dagster._core.remote_representation.external import (
        RemoteJob,
        RemoteRepository,
        RemoteResource,
    )
    from dagster._core.workspace.context import IWorkspaceProcessContext

TELEMETRY_STR = ".telemetry"
INSTANCE_ID_STR = "instance_id"
ENABLED_STR = "enabled"
UPDATE_REPO_STATS = "update_repo_stats"
# 'dagit' name is deprecated but we keep the same telemetry action name to avoid data disruption
START_DAGSTER_WEBSERVER = "start_dagit_webserver"
DAEMON_ALIVE = "daemon_alive"
SCHEDULED_RUN_CREATED = "scheduled_run_created"
SENSOR_RUN_CREATED = "sensor_run_created"
BACKFILL_RUN_CREATED = "backfill_run_created"
STEP_START_EVENT = "step_start_event"
STEP_SUCCESS_EVENT = "step_success_event"
STEP_FAILURE_EVENT = "step_failure_event"


TELEMETRY_WHITELISTED_FUNCTIONS = {
    "_logged_execute_job",
    "execute_execute_command",
    "execute_launch_command",
    "_daemon_run_command",
    "execute_materialize_command",
}


P = ParamSpec("P")
T = TypeVar("T")
T_Callable = TypeVar("T_Callable", bound=Callable[..., Any])


@overload
def telemetry_wrapper(target_fn: T_Callable) -> T_Callable: ...


@overload
def telemetry_wrapper(
    *,
    metadata: Mapping[str, str] | None,
) -> Callable[[Callable[P, T]], Callable[P, T]]: ...


def telemetry_wrapper(
    target_fn: T_Callable | None = None, *, metadata: Mapping[str, str] | None = None
) -> T_Callable | Callable[[Callable[P, T]], Callable[P, T]]:
    """Wrapper around functions that are logged. Will log the function_name, client_time, and
    elapsed_time, and success.

    Wrapped function must be in the list of whitelisted function, and must have a DagsterInstance
    parameter named 'instance' in the signature.
    """
    if target_fn is not None:
        return _telemetry_wrapper(target_fn)

    def _wraps(f: Callable[P, T]) -> Callable[P, T]:
        return _telemetry_wrapper(f, metadata)

    return _wraps


def _telemetry_wrapper(
    f: Callable[P, T], metadata: Mapping[str, str] | None = None
) -> Callable[P, T]:
    metadata = check.opt_mapping_param(metadata, "metadata", key_type=str, value_type=str)

    if f.__name__ not in TELEMETRY_WHITELISTED_FUNCTIONS:
        raise DagsterInvariantViolationError(
            f"Attempted to log telemetry for function {f.__name__} that is not in telemetry whitelisted "
            f"functions list: {TELEMETRY_WHITELISTED_FUNCTIONS}."
        )
    sig = inspect.signature(f)
    instance_index = None
    for i, name in enumerate(sig.parameters):
        if name == "instance":
            instance_index = i
            break

    if instance_index is None:
        raise DagsterInvariantViolationError(
            "Attempted to log telemetry for function {name} that does not take a DagsterInstance "
            "in a parameter called 'instance'"
        )

    @wraps(f)
    def wrap(*args: P.args, **kwargs: P.kwargs) -> T:
        instance = _check_telemetry_instance_param(args, kwargs, instance_index)
        start_time = datetime.datetime.now()
        log_action(
            instance=instance,
            action=f.__name__ + "_started",
            client_time=start_time,
            metadata=metadata,
        )
        result = f(*args, **kwargs)
        end_time = datetime.datetime.now()
        success_metadata = {"success": getattr(result, "success", None)}
        log_action(
            instance=instance,
            action=f.__name__ + "_ended",
            client_time=end_time,
            elapsed_time=end_time - start_time,
            metadata=merge_dicts(success_metadata, metadata),
        )
        return result

    return wrap


def get_log_queue_dir() -> str:
    """Get the directory where we store log queue files, creating the directory if needed.

    The log queue directory is used to temporarily store logs during upload. This is to prevent
    dropping events or double-sending events that occur during the upload process.

    If $DAGSTER_HOME is set, return $DAGSTER_HOME/.logs_queue/
    Otherwise, return ~/.dagster/.logs_queue/
    """
    dagster_home_path = dagster_home_if_set()
    if dagster_home_path is None:
        dagster_home_path = os.path.expanduser(DAGSTER_HOME_FALLBACK)

    dagster_home_logs_queue_path = dagster_home_path + "/.logs_queue/"
    if not os.path.exists(dagster_home_logs_queue_path):
        os.makedirs(dagster_home_logs_queue_path)

    return dagster_home_logs_queue_path


def _check_telemetry_instance_param(
    args: Sequence[object], kwargs: Mapping[str, object], instance_index: int
) -> DagsterInstance:
    if "instance" in kwargs:
        return check.inst_param(
            kwargs["instance"],  # type: ignore
            "instance",
            DagsterInstance,
            "'instance' parameter passed as keyword argument must be a DagsterInstance",
        )
    else:
        check.invariant(len(args) > instance_index)
        return check.inst_param(
            args[instance_index],  # type: ignore
            "instance",
            DagsterInstance,
            f"'instance' argument at position {instance_index} must be a DagsterInstance",
        )


def _get_instance_telemetry_info(
    instance: DagsterInstance,
) -> TelemetrySettings:
    from dagster._core.storage.runs import SqlRunStorage

    check.inst_param(instance, "instance", DagsterInstance)
    dagster_telemetry_enabled = _get_instance_telemetry_enabled(instance)
    instance_id = None
    run_storage_id = None
    if dagster_telemetry_enabled:
        instance_id = get_or_set_instance_id()
        if isinstance(instance.run_storage, SqlRunStorage):
            run_storage_id = instance.run_storage.get_run_storage_id()

    return TelemetrySettings(
        dagster_telemetry_enabled=dagster_telemetry_enabled,
        instance_id=instance_id,
        run_storage_id=run_storage_id,
    )


def _get_instance_telemetry_enabled(instance: DagsterInstance) -> bool:
    return instance.telemetry_enabled


def hash_name(name: str) -> str:
    return hashlib.sha256(name.encode("utf-8")).hexdigest()


def get_stats_from_remote_repo(remote_repo: "RemoteRepository") -> Mapping[str, str]:
    from dagster._core.definitions.partitions.snap import DynamicPartitionsSnap, MultiPartitionsSnap

    num_pipelines_in_repo = len(remote_repo.get_all_jobs())
    num_schedules_in_repo = len(remote_repo.get_schedules())
    num_sensors_in_repo = len(remote_repo.get_sensors())
    asset_node_snaps = remote_repo.get_asset_node_snaps()
    num_assets_in_repo = len(asset_node_snaps)
    remote_resources = remote_repo.get_resources()

    num_checks = len(remote_repo.repository_snap.asset_check_nodes or [])
    num_assets_with_checks = len(
        {c.asset_key for c in remote_repo.repository_snap.asset_check_nodes or []}
    )

    num_partitioned_assets_in_repo = 0
    num_multi_partitioned_assets_in_repo = 0
    num_dynamic_partitioned_assets_in_repo = 0
    num_assets_with_freshness_policies_in_repo = 0
    num_assets_with_eager_auto_materialize_policies_in_repo = 0
    num_assets_with_lazy_auto_materialize_policies_in_repo = 0
    num_assets_with_single_run_backfill_policies_in_repo = 0
    num_assets_with_multi_run_backfill_policies_in_repo = 0
    num_source_assets_in_repo = 0
    num_observable_source_assets_in_repo = 0
    num_dbt_assets_in_repo = 0
    num_assets_with_code_versions_in_repo = 0

    for asset in asset_node_snaps:
        if asset.partitions:
            num_partitioned_assets_in_repo += 1

            if isinstance(asset.partitions, DynamicPartitionsSnap):
                num_dynamic_partitioned_assets_in_repo += 1

            if isinstance(asset.partitions, MultiPartitionsSnap):
                num_multi_partitioned_assets_in_repo += 1

        if asset.legacy_freshness_policy is not None:
            num_assets_with_freshness_policies_in_repo += 1

        if asset.auto_materialize_policy is not None:
            if asset.auto_materialize_policy.policy_type == AutoMaterializePolicyType.EAGER:
                num_assets_with_eager_auto_materialize_policies_in_repo += 1
            else:
                num_assets_with_lazy_auto_materialize_policies_in_repo += 1

        if asset.backfill_policy is not None:
            if asset.backfill_policy.policy_type == BackfillPolicyType.SINGLE_RUN:
                num_assets_with_single_run_backfill_policies_in_repo += 1
            else:
                num_assets_with_multi_run_backfill_policies_in_repo += 1

        if asset.is_source:
            num_source_assets_in_repo += 1

            if asset.is_observable:
                num_observable_source_assets_in_repo += 1

        if asset.code_version is not None:
            num_assets_with_code_versions_in_repo += 1

        if asset.compute_kind == "dbt":
            num_dbt_assets_in_repo += 1

    num_asset_reconciliation_sensors_in_repo = sum(
        1 for sensor in remote_repo.get_sensors() if sensor.name == "asset_reconciliation_sensor"
    )

    return {
        **get_resource_stats(remote_resources=list(remote_resources)),
        "num_pipelines_in_repo": str(num_pipelines_in_repo),
        "num_schedules_in_repo": str(num_schedules_in_repo),
        "num_sensors_in_repo": str(num_sensors_in_repo),
        "num_assets_in_repo": str(num_assets_in_repo),
        "num_source_assets_in_repo": str(num_source_assets_in_repo),
        "num_partitioned_assets_in_repo": str(num_partitioned_assets_in_repo),
        "num_dynamic_partitioned_assets_in_repo": str(num_dynamic_partitioned_assets_in_repo),
        "num_multi_partitioned_assets_in_repo": str(num_multi_partitioned_assets_in_repo),
        "num_assets_with_freshness_policies_in_repo": str(
            num_assets_with_freshness_policies_in_repo
        ),
        "num_assets_with_eager_auto_materialize_policies_in_repo": str(
            num_assets_with_eager_auto_materialize_policies_in_repo
        ),
        "num_assets_with_lazy_auto_materialize_policies_in_repo": str(
            num_assets_with_lazy_auto_materialize_policies_in_repo
        ),
        "num_assets_with_single_run_backfill_policies_in_repo": str(
            num_assets_with_single_run_backfill_policies_in_repo
        ),
        "num_assets_with_multi_run_backfill_policies_in_repo": str(
            num_assets_with_multi_run_backfill_policies_in_repo
        ),
        "num_observable_source_assets_in_repo": str(num_observable_source_assets_in_repo),
        "num_dbt_assets_in_repo": str(num_dbt_assets_in_repo),
        "num_assets_with_code_versions_in_repo": str(num_assets_with_code_versions_in_repo),
        "num_asset_reconciliation_sensors_in_repo": str(num_asset_reconciliation_sensors_in_repo),
        "num_asset_checks": str(num_checks),
        "num_assets_with_checks": str(num_assets_with_checks),
    }


def get_resource_stats(remote_resources: Sequence["RemoteResource"]) -> Mapping[str, Any]:
    used_dagster_resources = []
    used_custom_resources = False

    for resource in remote_resources:
        resource_type = resource.resource_type
        split_resource_type = resource_type.split(".")
        module_name = split_resource_type[0]
        class_name = split_resource_type[-1]

        if resource.is_dagster_maintained:
            used_dagster_resources.append({"module_name": module_name, "class_name": class_name})
        else:
            used_custom_resources = True

    return {
        "dagster_resources": used_dagster_resources,
        "has_custom_resources": str(used_custom_resources),
    }


def log_remote_repo_stats(
    instance: DagsterInstance,
    source: str,
    remote_repo: "RemoteRepository",
    remote_job: Optional["RemoteJob"] = None,
):
    from dagster._core.remote_representation.external import RemoteJob, RemoteRepository

    check.inst_param(instance, "instance", DagsterInstance)
    check.str_param(source, "source")
    check.inst_param(remote_repo, "remote_repo", RemoteRepository)
    check.opt_inst_param(remote_job, "remote_job", RemoteJob)

    if _get_instance_telemetry_enabled(instance):
        instance_id = get_or_set_instance_id()

        job_name_hash = hash_name(remote_job.name) if remote_job else ""
        repo_hash = hash_name(remote_repo.name)
        location_name_hash = hash_name(remote_repo.handle.location_name)

        write_telemetry_log_line(
            TelemetryEntry(
                action=UPDATE_REPO_STATS,
                client_time=str(datetime.datetime.now()),
                event_id=str(uuid.uuid4()),
                instance_id=instance_id,
                user_id=get_or_set_user_id(),
                metadata={
                    **get_stats_from_remote_repo(remote_repo),
                    "source": source,
                    "pipeline_name_hash": job_name_hash,
                    "repo_hash": repo_hash,
                    "location_name_hash": location_name_hash,
                },
            )._asdict()
        )


def log_repo_stats(
    instance: DagsterInstance,
    source: str,
    job: IJob | None = None,
    repo: ReconstructableRepository | None = None,
) -> None:
    from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
    from dagster._core.definitions.partitions.definition import DynamicPartitionsDefinition

    check.inst_param(instance, "instance", DagsterInstance)
    check.str_param(source, "source")
    check.opt_inst_param(job, "job", IJob)
    check.opt_inst_param(repo, "repo", ReconstructableRepository)

    def _get_num_dynamic_partitioned_assets(asset_defs: Sequence[AssetsDefinition]) -> int:
        return sum(
            1
            for asset in asset_defs
            if asset.partitions_def
            and isinstance(asset.partitions_def, DynamicPartitionsDefinition)
        )

    if _get_instance_telemetry_enabled(instance):
        instance_id = get_or_set_instance_id()

        if isinstance(job, ReconstructableJob):
            job_name_hash = hash_name(job.get_definition().name)
            repository = job.get_reconstructable_repository().get_definition()
            repo_hash = hash_name(repository.name)
            num_jobs_in_repo = len(repository.job_names)
            num_schedules_in_repo = len(repository.schedule_defs)
            num_sensors_in_repo = len(repository.sensor_defs)
            all_assets = list(repository.assets_defs_by_key.values())
            num_assets_in_repo = len(all_assets)
            num_dynamic_partitioned_assets_in_repo = _get_num_dynamic_partitioned_assets(all_assets)
        elif isinstance(repo, ReconstructableRepository):
            job_name_hash = ""
            repository = repo.get_definition()
            repo_hash = hash_name(repository.name)
            num_jobs_in_repo = len(repository.job_names)
            num_schedules_in_repo = len(repository.schedule_defs)
            num_sensors_in_repo = len(repository.sensor_defs)
            all_assets = list(repository.assets_defs_by_key.values())
            num_assets_in_repo = len(all_assets)
            num_dynamic_partitioned_assets_in_repo = _get_num_dynamic_partitioned_assets(all_assets)
        else:
            job_name_hash = hash_name(job.get_definition().name)  # type: ignore
            repo_hash = hash_name(get_ephemeral_repository_name(job.get_definition().name))  # type: ignore
            num_jobs_in_repo = 1
            num_schedules_in_repo = 0
            num_sensors_in_repo = 0
            num_assets_in_repo = 0
            num_dynamic_partitioned_assets_in_repo = 0

        write_telemetry_log_line(
            TelemetryEntry(
                action=UPDATE_REPO_STATS,
                client_time=str(datetime.datetime.now()),
                event_id=str(uuid.uuid4()),
                instance_id=instance_id,
                user_id=get_or_set_user_id(),
                metadata={
                    "source": source,
                    "pipeline_name_hash": job_name_hash,
                    "num_pipelines_in_repo": str(num_jobs_in_repo),
                    "num_schedules_in_repo": str(num_schedules_in_repo),
                    "num_sensors_in_repo": str(num_sensors_in_repo),
                    "num_assets_in_repo": str(num_assets_in_repo),
                    "repo_hash": repo_hash,
                    "num_dynamic_partitioned_assets_in_repo": str(
                        num_dynamic_partitioned_assets_in_repo
                    ),
                },
            )._asdict()
        )


def log_workspace_stats(
    instance: DagsterInstance, workspace_process_context: "IWorkspaceProcessContext"
) -> None:
    from dagster._core.workspace.context import IWorkspaceProcessContext

    check.inst_param(instance, "instance", DagsterInstance)
    check.inst_param(
        workspace_process_context, "workspace_process_context", IWorkspaceProcessContext
    )

    request_context = workspace_process_context.create_request_context()

    for code_location in request_context.code_locations:
        for repo in code_location.get_repositories().values():
            log_remote_repo_stats(instance, source="dagit", remote_repo=repo)


def log_action(
    instance: DagsterInstance,
    action: str,
    client_time: datetime.datetime | None = None,
    elapsed_time: datetime.timedelta | None = None,
    metadata: Mapping[str, str] | None = None,
) -> None:
    log_telemetry_action(
        lambda: _get_instance_telemetry_info(instance),
        action,
        client_time,
        elapsed_time,
        metadata,
    )


def log_dagster_event(event: DagsterEvent, job_context: PlanOrchestrationContext) -> None:
    if not any((event.is_step_start, event.is_step_success, event.is_step_failure)):
        return

    metadata = {
        "run_id_hash": hash_name(job_context.run_id),
        "step_key_hash": hash_name(event.step_key),  # type: ignore
    }

    if event.is_step_start:
        action = STEP_START_EVENT
    elif event.is_step_success:
        action = STEP_SUCCESS_EVENT
        if isinstance(event.event_specific_data, StepSuccessData):  # make mypy happy
            metadata["duration_ms"] = event.event_specific_data.duration_ms  # type: ignore
    else:  # event.is_step_failure
        action = STEP_FAILURE_EVENT

    log_action(
        instance=job_context.instance,
        action=action,
        client_time=datetime.datetime.now(),
        metadata=metadata,
    )



from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional, TypeAlias

from dagster_shared import record
from dagster_shared.serdes import whitelist_for_serdes

import dagster._check as check
from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.asset_health.asset_health import AssetHealthStatus
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition import PartitionsDefinition
from dagster._core.definitions.partitions.snap import PartitionsSnap
from dagster._core.loader import LoadableBy, LoadingContext
from dagster._core.storage.dagster_run import RunRecord
from dagster._core.storage.event_log.base import AssetRecord
from dagster._core.storage.partition_status_cache import get_partition_subsets

if TYPE_CHECKING:
    from dagster._core.workspace.context import BaseWorkspaceRequestContext


@whitelist_for_serdes
@record.record
class MinimalAssetMaterializationHealthState(LoadableBy[AssetKey]):
    """Minimal object for computing the health status for the materialization state of an asset.
    This object is intended to be small and quick to deserialize. Deserializing AssetMaterializationHealthState
    can be slow if there is a large entity subset. Rather than storing entity subsets, we store the number
    of partitions in each state. This lets us quickly compute the health status of the asset and create
    the metadata required for the UI.
    """

    latest_materialization_timestamp: float | None
    latest_terminal_run_id: str | None
    num_failed_partitions: int
    num_currently_materialized_partitions: int
    partitions_snap: PartitionsSnap | None
    latest_failed_to_materialize_timestamp: float | None = None
    latest_failed_to_materialize_run_id: str | None = None

    @property
    def health_status(self) -> AssetHealthStatus:
        if self.num_failed_partitions == 0 and self.num_currently_materialized_partitions == 0:
            return AssetHealthStatus.UNKNOWN
        if self.num_failed_partitions > 0:
            return AssetHealthStatus.DEGRADED
        else:
            return AssetHealthStatus.HEALTHY

    @property
    def partitions_def(self) -> PartitionsDefinition | None:
        if self.partitions_snap is None:
            return None
        return self.partitions_snap.get_partitions_definition()

    @classmethod
    def from_asset_materialization_health_state(
        cls,
        asset_materialization_health_state: "AssetMaterializationHealthState",
    ) -> "MinimalAssetMaterializationHealthState":
        return cls(
            latest_materialization_timestamp=asset_materialization_health_state.latest_materialization_timestamp,
            latest_terminal_run_id=asset_materialization_health_state.latest_terminal_run_id,
            num_failed_partitions=asset_materialization_health_state.failed_subset.size,
            num_currently_materialized_partitions=asset_materialization_health_state.currently_materialized_subset.size,
            partitions_snap=asset_materialization_health_state.partitions_snap,
            latest_failed_to_materialize_timestamp=asset_materialization_health_state.latest_failed_to_materialize_timestamp,
            latest_failed_to_materialize_run_id=asset_materialization_health_state.latest_failed_to_materialize_run_id,
        )

    @classmethod
    def _blocking_batch_load(
        cls, keys: Iterable[AssetKey], context: LoadingContext
    ) -> Iterable[Optional["MinimalAssetMaterializationHealthState"]]:
        asset_materialization_health_states = (
            context.instance.get_minimal_asset_materialization_health_state_for_assets(list(keys))
        )
        return [asset_materialization_health_states.get(key) for key in keys]


@whitelist_for_serdes
@record.record
class AssetMaterializationHealthState(LoadableBy[AssetKey]):
    """For tracking the materialization health of an asset, we only care about the most recent
    completed materialization attempt for each asset/partition. This record keeps track of the
    assets/partitions that have ever been successfully materialized and those that are currently in
    a failed state. From this information we can derive the subset that is currently in a successfully
    materialized state.

    If an asset/partition is currently being materialized, it will not move to a new state until after
    the materialization attempt is complete.

    In the future, we may want to expand this to track the last N materialization successes/failures for
    each asset. We could also maintain a list of in progress materializations, but that requires streamline to be
    better able to handle runs being deleted.

    materialized_subset: The subset of the asset that has ever been successfully materialized.
    failed_subset: The subset of the asset that is currently in a failed state.
    partitions_snap: The partitions definition for the asset. None if it is not a partitioned asset.
    latest_terminal_run_id: The id of the latest run with a successful or failed materialization event for the asset.
    """

    materialized_subset: SerializableEntitySubset[AssetKey]
    failed_subset: SerializableEntitySubset[AssetKey]
    partitions_snap: PartitionsSnap | None
    latest_terminal_run_id: str | None
    latest_materialization_timestamp: float | None = None
    latest_failed_to_materialize_timestamp: float | None = None
    latest_failed_to_materialize_run_id: str | None = None

    @property
    def partitions_def(self) -> PartitionsDefinition | None:
        if self.partitions_snap is None:
            return None
        return self.partitions_snap.get_partitions_definition()

    @property
    def currently_materialized_subset(self) -> SerializableEntitySubset[AssetKey]:
        """The subset of the asset that is currently in a successfully materialized state."""
        return self.materialized_subset.compute_difference(self.failed_subset)

    @property
    def health_status(self) -> AssetHealthStatus:
        if self.materialized_subset.is_empty and self.failed_subset.is_empty:
            return AssetHealthStatus.UNKNOWN
        elif not self.failed_subset.is_empty:
            return AssetHealthStatus.DEGRADED
        else:
            return AssetHealthStatus.HEALTHY

    @classmethod
    async def compute_for_asset(
        cls,
        asset_key: AssetKey,
        partitions_def: PartitionsDefinition | None,
        loading_context: LoadingContext,
    ) -> "AssetMaterializationHealthState":
        """Creates an AssetMaterializationHealthState for the given asset. Requires fetching the AssetRecord
        and potentially the latest run from the DB, or regenerating the partition status cache.
        """
        asset_record = await AssetRecord.gen(loading_context, asset_key)

        if partitions_def is not None:
            (
                materialized_partition_subset,
                failed_partition_subset,
                _,
            ) = await get_partition_subsets(
                loading_context.instance,
                loading_context,
                asset_key,
                loading_context.instance,
                partitions_def,
            )

            if materialized_partition_subset is None or failed_partition_subset is None:
                check.failed("Expected partitions subset for a partitioned asset")

            last_run_id = None
            last_failed_run_id = None
            latest_materialization_timestamp = None
            latest_failed_to_materialize_timestamp = None
            if asset_record is not None:
                entry = asset_record.asset_entry
                latest_record = max(
                    [
                        entry.last_materialization_record,
                        entry.last_failed_to_materialize_record,
                    ],
                    key=lambda record: -1 if record is None else record.storage_id,
                )
                last_run_id = latest_record.run_id if latest_record else None
                last_failed_run_id = (
                    entry.last_failed_to_materialize_record.run_id
                    if entry.last_failed_to_materialize_record
                    else None
                )
                latest_materialization_timestamp = (
                    entry.last_materialization_record.timestamp
                    if entry.last_materialization_record
                    else None
                )
                latest_failed_to_materialize_timestamp = (
                    entry.last_failed_to_materialize_record.timestamp
                    if entry.last_failed_to_materialize_record
                    else None
                )

            return cls(
                materialized_subset=SerializableEntitySubset(
                    key=asset_key, value=materialized_partition_subset
                ),
                failed_subset=SerializableEntitySubset(
                    key=asset_key, value=failed_partition_subset
                ),
                partitions_snap=PartitionsSnap.from_def(partitions_def),
                latest_terminal_run_id=last_run_id,
                latest_materialization_timestamp=latest_materialization_timestamp,
                latest_failed_to_materialize_timestamp=latest_failed_to_materialize_timestamp,
                latest_failed_to_materialize_run_id=last_failed_run_id,
            )

        if asset_record is None:
            return AssetMaterializationHealthState(
                materialized_subset=SerializableEntitySubset(key=asset_key, value=False),
                failed_subset=SerializableEntitySubset(key=asset_key, value=False),
                partitions_snap=None,
                latest_terminal_run_id=None,
                latest_materialization_timestamp=None,
                latest_failed_to_materialize_timestamp=None,
                latest_failed_to_materialize_run_id=None,
            )

        asset_entry = asset_record.asset_entry
        latest_materialization_timestamp = (
            asset_entry.last_materialization_record.timestamp
            if asset_entry.last_materialization_record
            else None
        )
        latest_failed_to_materialize_timestamp = (
            asset_entry.last_failed_to_materialize_record.timestamp
            if asset_entry.last_failed_to_materialize_record
            else None
        )
        latest_failed_run_id = (
            asset_entry.last_failed_to_materialize_record.run_id
            if asset_entry.last_failed_to_materialize_record
            else None
        )
        if asset_entry.last_run_id is None:
            return AssetMaterializationHealthState(
                materialized_subset=SerializableEntitySubset(key=asset_key, value=False),
                failed_subset=SerializableEntitySubset(key=asset_key, value=False),
                partitions_snap=None,
                latest_terminal_run_id=None,
                latest_materialization_timestamp=latest_materialization_timestamp,
                latest_failed_to_materialize_timestamp=latest_failed_to_materialize_timestamp,
                latest_failed_to_materialize_run_id=latest_failed_run_id,
            )

        has_ever_materialized = asset_entry.last_materialization is not None
        (
            is_currently_failed,
            latest_terminal_run_id,
        ) = await _get_is_currently_failed_and_latest_terminal_run_id(loading_context, asset_record)

        return cls(
            materialized_subset=SerializableEntitySubset(
                key=asset_key, value=has_ever_materialized
            ),
            failed_subset=SerializableEntitySubset(key=asset_key, value=is_currently_failed),
            partitions_snap=None,
            latest_terminal_run_id=latest_terminal_run_id,
            latest_materialization_timestamp=latest_materialization_timestamp,
            latest_failed_to_materialize_timestamp=latest_failed_to_materialize_timestamp,
            latest_failed_to_materialize_run_id=latest_failed_run_id,
        )

    @classmethod
    def _blocking_batch_load(
        cls, keys: Iterable[AssetKey], context: LoadingContext
    ) -> Iterable[Optional["AssetMaterializationHealthState"]]:
        asset_materialization_health_states = (
            context.instance.get_asset_materialization_health_state_for_assets(list(keys))
        )
        return [asset_materialization_health_states.get(key) for key in keys]


async def _get_is_currently_failed_and_latest_terminal_run_id(
    loading_context: LoadingContext, asset_record: AssetRecord
) -> tuple[bool, str | None]:
    """Determines if the asset is currently in a failed state. If we are storing failure events for the
    asset, this can be determined by looking at the AssetRecord. For assets where we are not storing failure
    events, we have to derive the failure state from the latest run record.

    Also returns the id of the latest run with a successful or failed materialization event for the asset.
    """
    asset_entry = asset_record.asset_entry
    if loading_context.instance.can_read_failure_events_for_asset(asset_record):
        latest_record = max(
            [
                asset_entry.last_materialization_record,
                asset_entry.last_failed_to_materialize_record,
            ],
            key=lambda record: -1 if record is None else record.storage_id,
        )
        return (
            latest_record.storage_id == asset_entry.last_failed_to_materialize_storage_id
            if latest_record
            else False,
            latest_record.run_id if latest_record else None,
        )

    if asset_entry.last_run_id is None:
        return False, None

    # if failure events are not stored, we usually have to fetch the run record to check if the
    # asset is currently failed. However, if the latest run id is the same as the last materialization run id,
    # then we know the asset is in a successfully materialized state.
    if (
        asset_entry.last_materialization
        and asset_entry.last_run_id == asset_entry.last_materialization.run_id
    ):
        return False, asset_entry.last_materialization.run_id

    run_record = await RunRecord.gen(loading_context, asset_entry.last_run_id)
    if run_record is None or not run_record.dagster_run.is_finished or run_record.end_time is None:
        # the run is deleted or in progress. With the information we have available, we cannot know
        # if the asset is in a failed state prior to this run. Historically, we have resorted to
        # reporting the asset as materialized if it has ever been materialized, and otherwise report it
        # as not materialized.
        return (
            False,
            asset_entry.last_materialization.run_id if asset_entry.last_materialization else None,
        )

    if (
        asset_entry.last_materialization
        and asset_entry.last_materialization.timestamp > run_record.end_time
    ):
        # the latest materialization was reported manually
        return False, asset_entry.last_materialization.run_id

    # if the run failed, then report the asset as failed
    return run_record.dagster_run.is_failure, run_record.dagster_run.run_id


@whitelist_for_serdes
@record.record
class AssetHealthMaterializationDegradedPartitionedMeta:
    num_failed_partitions: int
    num_missing_partitions: int
    total_num_partitions: int
    latest_run_id: str | None = None
    latest_failed_to_materialize_run_id: str | None = None


@whitelist_for_serdes
@record.record
class AssetHealthMaterializationHealthyPartitionedMeta:
    num_missing_partitions: int
    total_num_partitions: int
    latest_run_id: str | None = None
    latest_failed_to_materialize_run_id: str | None = None


@whitelist_for_serdes
@record.record
class AssetHealthMaterializationDegradedNotPartitionedMeta:
    failed_run_id: str | None


AssetHealthMaterializationMetadata: TypeAlias = (
    AssetHealthMaterializationDegradedPartitionedMeta
    | AssetHealthMaterializationHealthyPartitionedMeta
    | AssetHealthMaterializationDegradedNotPartitionedMeta
)


async def get_materialization_status_and_metadata(
    context: "BaseWorkspaceRequestContext", asset_key: AssetKey
) -> tuple[AssetHealthStatus, Optional["AssetHealthMaterializationMetadata"]]:
    """Gets an AssetMaterializationHealthState object for an asset, either via streamline or by computing
    it based on the state of the DB. Then converts it to a AssetHealthStatus and the metadata
    needed to power the UIs. Metadata is fetched from the AssetLatestMaterializationState object, again
    either via streamline or by computing it based on the state of the DB.
    """
    asset_materialization_health_state = await MinimalAssetMaterializationHealthState.gen(
        context, asset_key
    )
    if asset_materialization_health_state is None:
        # if the minimal health stat does not exist, try fetching the full health state. It's possible that
        # deserializing the full health state is non-performant since it contains a serialized entity subset, which
        # is why we only fetch it if the minimal health state does not exist.
        slow_deserialize_asset_materialization_health_state = (
            await AssetMaterializationHealthState.gen(context, asset_key)
        )
        if slow_deserialize_asset_materialization_health_state is not None:
            asset_materialization_health_state = (
                MinimalAssetMaterializationHealthState.from_asset_materialization_health_state(
                    slow_deserialize_asset_materialization_health_state
                )
            )
    # captures streamline disabled or consumer state doesn't exist
    if asset_materialization_health_state is None:
        if context.instance.streamline_read_asset_health_required("asset-materialization-health"):
            return AssetHealthStatus.UNKNOWN, None

        if not context.asset_graph.has(asset_key):
            # if the asset is not in the asset graph, it could be because materializations are reported by
            # an external system, determine the status as best we can based on the asset record
            asset_record = await AssetRecord.gen(context, asset_key)
            if asset_record is None:
                return AssetHealthStatus.UNKNOWN, None
            has_ever_materialized = asset_record.asset_entry.last_materialization is not None
            is_currently_failed, run_id = await _get_is_currently_failed_and_latest_terminal_run_id(
                context, asset_record
            )
            if is_currently_failed:
                meta = AssetHealthMaterializationDegradedNotPartitionedMeta(
                    failed_run_id=run_id,
                )
                return AssetHealthStatus.DEGRADED, meta
            if has_ever_materialized:
                return AssetHealthStatus.HEALTHY, None
            else:
                if asset_record.asset_entry.last_observation is not None:
                    return AssetHealthStatus.HEALTHY, None
                return AssetHealthStatus.UNKNOWN, None

        node_snap = context.asset_graph.get(asset_key)
        if node_snap.is_observable and not node_snap.is_materializable:  # observable source asset
            # get the asset record to see if there is an observation event
            asset_record = await AssetRecord.gen(context, asset_key)
            if asset_record and asset_record.asset_entry.last_observation is not None:
                return AssetHealthStatus.HEALTHY, None
            return AssetHealthStatus.UNKNOWN, None

        asset_materialization_health_state = (
            MinimalAssetMaterializationHealthState.from_asset_materialization_health_state(
                await AssetMaterializationHealthState.compute_for_asset(
                    asset_key,
                    node_snap.partitions_def,
                    context,
                )
            )
        )

    if asset_materialization_health_state.health_status == AssetHealthStatus.HEALTHY:
        num_missing = 0
        total_num_partitions = 0
        if asset_materialization_health_state.partitions_def is not None:
            with partition_loading_context(dynamic_partitions_store=context.instance):
                total_num_partitions = (
                    asset_materialization_health_state.partitions_def.get_num_partitions()
                )
            # asset is healthy, so no partitions are failed
            num_missing = (
                total_num_partitions
                - asset_materialization_health_state.num_currently_materialized_partitions
            )
        if num_missing > 0 and total_num_partitions > 0:
            meta = AssetHealthMaterializationHealthyPartitionedMeta(
                num_missing_partitions=num_missing,
                total_num_partitions=total_num_partitions,
                latest_run_id=asset_materialization_health_state.latest_terminal_run_id,
                latest_failed_to_materialize_run_id=asset_materialization_health_state.latest_failed_to_materialize_run_id,
            )
        else:
            # captures the case when asset is not partitioned, or the asset is partitioned and all partitions are materialized
            meta = None
        return AssetHealthStatus.HEALTHY, meta
    elif asset_materialization_health_state.health_status == AssetHealthStatus.DEGRADED:
        if asset_materialization_health_state.partitions_def is not None:
            with partition_loading_context(dynamic_partitions_store=context.instance):
                total_num_partitions = (
                    asset_materialization_health_state.partitions_def.get_num_partitions()
                )
            num_missing = (
                total_num_partitions
                - asset_materialization_health_state.num_currently_materialized_partitions
                - asset_materialization_health_state.num_failed_partitions
            )
            meta = AssetHealthMaterializationDegradedPartitionedMeta(
                num_failed_partitions=asset_materialization_health_state.num_failed_partitions,
                num_missing_partitions=num_missing,
                total_num_partitions=total_num_partitions,
                latest_run_id=asset_materialization_health_state.latest_terminal_run_id,
                latest_failed_to_materialize_run_id=asset_materialization_health_state.latest_failed_to_materialize_run_id,
            )
        else:
            meta = AssetHealthMaterializationDegradedNotPartitionedMeta(
                failed_run_id=asset_materialization_health_state.latest_failed_to_materialize_run_id,
            )
        return AssetHealthStatus.DEGRADED, meta
    elif asset_materialization_health_state.health_status == AssetHealthStatus.UNKNOWN:
        return AssetHealthStatus.UNKNOWN, None
    else:
        check.failed(
            f"Unexpected materialization health status: {asset_materialization_health_state.health_status}"
        )



import textwrap

from pydantic import BaseModel

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.decorators.asset_check_decorator import asset_check
from dagster.components.lib.shim_components.base import ShimScaffolder
from dagster.components.scaffold.scaffold import ScaffoldRequest, scaffold_with


class AssetCheckScaffoldParams(BaseModel):
    """Parameters for the AssetCheckScaffolder."""

    asset_key: str


class AssetCheckScaffolder(ShimScaffolder[AssetCheckScaffoldParams]):
    @classmethod
    def get_scaffold_params(cls) -> type[AssetCheckScaffoldParams]:
        return AssetCheckScaffoldParams

    def get_text(self, request: ScaffoldRequest[AssetCheckScaffoldParams]) -> str:
        """Get the text for an asset check.

        Args:
            request: The scaffold request containing type name, target path, format, project root and optional params

        Returns:
            The text for the asset check
        """
        assert request.params
        asset_key = AssetKey.from_user_string(request.params.asset_key)
        return textwrap.dedent(
            f"""\
            import dagster as dg


            @dg.asset_check(asset=dg.AssetKey({asset_key.path!s}))
            def {request.target_path.stem}(context: dg.AssetCheckExecutionContext) -> dg.AssetCheckResult: ...
            """
        )


scaffold_with(AssetCheckScaffolder)(asset_check)



from collections.abc import Mapping

import dagster_shared.check as check
from dagster_shared.record import IHaveNew, record, record_custom, replace

from dagster._core.definitions.asset_checks.asset_check_spec import (
    AssetCheckKey,
    AssetCheckSeverity,
)
from dagster._core.definitions.events import AssetKey, MetadataValue, RawMetadataValue
from dagster._core.definitions.metadata import normalize_metadata
from dagster._core.definitions.partitions.subset import PartitionsSubset
from dagster._serdes import whitelist_for_serdes


@whitelist_for_serdes
@record
class AssetCheckEvaluationPlanned:
    """Metadata for the event when an asset check is launched."""

    asset_key: AssetKey
    check_name: str
    partitions_subset: PartitionsSubset | None = None

    @property
    def asset_check_key(self) -> AssetCheckKey:
        return AssetCheckKey(self.asset_key, self.check_name)


@whitelist_for_serdes
@record
class AssetCheckEvaluationTargetMaterializationData:
    """A pointer to the latest materialization at execution time of an asset check."""

    storage_id: int
    run_id: str
    timestamp: float


@whitelist_for_serdes(storage_field_names={"passed": "success"})
@record_custom
class AssetCheckEvaluation(IHaveNew):
    """Represents the outcome of a evaluating an asset check.

    Args:
        asset_key (AssetKey):
            The asset key that was checked.
        check_name (str):
            The name of the check.
        passed (bool):
            The pass/fail result of the check.
        metadata (Optional[Mapping[str, MetadataValue]]):
            Arbitrary user-provided metadata about the asset.  Keys are displayed string labels, and
            values are one of the following: string, float, int, JSON-serializable dict, JSON-serializable
            list, and one of the data classes returned by a MetadataValue static method.
        target_materialization_data (Optional[AssetCheckEvaluationTargetMaterializationData]):
            The latest materialization at execution time of the check.
        severity (AssetCheckSeverity):
            Severity of the check result.
        description (Optional[str]):
            A text description of the result of the check evaluation.
        blocking (Optional[bool]):
            Whether the check is blocking.
        partition (Optional[str]):
            The partition that the check was evaluated on, if applicable.
    """

    asset_key: AssetKey
    check_name: str
    passed: bool
    metadata: Mapping[str, MetadataValue]
    target_materialization_data: AssetCheckEvaluationTargetMaterializationData | None
    severity: AssetCheckSeverity
    description: str | None
    blocking: bool | None
    partition: str | None

    def __new__(
        cls,
        asset_key: AssetKey,
        check_name: str,
        passed: bool,
        metadata: Mapping[str, RawMetadataValue] | None = None,
        target_materialization_data: AssetCheckEvaluationTargetMaterializationData | None = None,
        severity: AssetCheckSeverity = AssetCheckSeverity.ERROR,
        description: str | None = None,
        blocking: bool | None = None,
        partition: str | None = None,
    ):
        return super().__new__(
            cls,
            asset_key=asset_key,
            check_name=check_name,
            passed=passed,
            metadata=normalize_metadata(
                check.opt_mapping_param(metadata, "metadata", key_type=str)
            ),
            target_materialization_data=target_materialization_data,
            severity=severity,
            description=description,
            blocking=blocking,
            partition=partition,
        )

    @property
    def asset_check_key(self) -> AssetCheckKey:
        return AssetCheckKey(self.asset_key, self.check_name)

    def with_metadata(self, metadata: Mapping[str, RawMetadataValue]) -> "AssetCheckEvaluation":
        normed_metadata = normalize_metadata(
            check.opt_mapping_param(metadata, "metadata", key_type=str),
        )
        return replace(self, metadata=normed_metadata)



from collections.abc import Callable, Mapping, Sequence
from datetime import datetime

import dagster._check as check
from dagster._annotations import deprecated_param, public
from dagster._core.definitions.partitions.definition.dynamic import DynamicPartitionsDefinition
from dagster._core.definitions.partitions.partitioned_config.partitioned_config import (
    PartitionConfigFn,
    PartitionedConfig,
)
from dagster._utils.warnings import normalize_renamed_param


@deprecated_param(
    param="tags_for_partition_fn",
    breaking_version="2.0",
    additional_warn_text="Use tags_for_partition_key_fn instead.",
)
@public
def dynamic_partitioned_config(
    partition_fn: Callable[[datetime | None], Sequence[str]],
    tags_for_partition_fn: Callable[[str], Mapping[str, str]] | None = None,
    tags_for_partition_key_fn: Callable[[str], Mapping[str, str]] | None = None,
) -> Callable[[PartitionConfigFn], PartitionedConfig]:
    """Creates a dynamic partitioned config for a job.

    The provided partition_fn returns a list of strings identifying the set of partitions, given
    an optional datetime argument (representing the current time).  The list of partitions returned
    may change over time.

    The decorated function takes in a partition key and returns a valid run config for a particular
    target job.

    Args:
        partition_fn (Callable[[datetime.datetime], Sequence[str]]): A function that generates a
            list of valid partition keys, which serve as the range of values that can be provided
            to the decorated run config function.
        tags_for_partition_fn (Optional[Callable[[str], Mapping[str, str]]]): A function that
            accepts a partition key and returns a dictionary of tags to attach to runs for that
            partition.

    Returns:
        PartitionedConfig
    """
    check.callable_param(partition_fn, "partition_fn")

    tags_for_partition_key_fn = normalize_renamed_param(
        tags_for_partition_key_fn,
        "tags_for_partition_key_fn",
        tags_for_partition_fn,
        "tags_for_partition_fn",
    )

    def inner(fn: PartitionConfigFn) -> PartitionedConfig:
        return PartitionedConfig(
            partitions_def=DynamicPartitionsDefinition(partition_fn),
            run_config_for_partition_key_fn=fn,
            decorated_fn=fn,
            tags_for_partition_key_fn=tags_for_partition_key_fn,
        )

    return inner



from dagster._core.definitions.partitions.subset.all import (
    AllPartitionsSubset as AllPartitionsSubset,
)
from dagster._core.definitions.partitions.subset.default import (
    DefaultPartitionsSubset as DefaultPartitionsSubset,
)
from dagster._core.definitions.partitions.subset.key_ranges import (
    KeyRangesPartitionsSubset as KeyRangesPartitionsSubset,
)
from dagster._core.definitions.partitions.subset.partitions_subset import (
    PartitionsSubset as PartitionsSubset,
)
from dagster._core.definitions.partitions.subset.time_window import (
    TimeWindowPartitionsSubset as TimeWindowPartitionsSubset,
)



from collections.abc import Generator, Mapping, Sequence
from contextlib import contextmanager
from typing import IO, Any

from typing_extensions import Self

import dagster._check as check
from dagster._annotations import public
from dagster._core.storage.compute_log_manager import (
    CapturedLogContext,
    CapturedLogMetadata,
    CapturedLogSubscription,
    ComputeIOType,
    ComputeLogManager,
)
from dagster._serdes import ConfigurableClass, ConfigurableClassData


@public
class NoOpComputeLogManager(ComputeLogManager, ConfigurableClass):
    """When enabled for a Dagster instance, stdout and stderr will not be available for any step."""

    def __init__(self, inst_data: ConfigurableClassData | None = None):
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @classmethod
    def from_config_value(
        cls, inst_data: ConfigurableClassData, config_value: Mapping[str, Any]
    ) -> Self:
        return cls(inst_data=inst_data, **config_value)

    @contextmanager
    def capture_logs(self, log_key: Sequence[str]) -> Generator[CapturedLogContext, None, None]:
        yield CapturedLogContext(log_key=log_key)

    def is_capture_complete(self, log_key: Sequence[str]):
        return True

    @contextmanager
    def open_log_stream(
        self, log_key: Sequence[str], io_type: ComputeIOType
    ) -> Generator[IO | None, None, None]:
        yield None

    def get_log_data_for_type(
        self,
        log_key: Sequence[str],
        io_type: ComputeIOType,
        offset: int,
        max_bytes: int | None,
    ) -> tuple[bytes | None, int]:
        return None, 0

    def get_log_metadata(self, log_key: Sequence[str]) -> CapturedLogMetadata:
        return CapturedLogMetadata()

    def delete_logs(
        self, log_key: Sequence[str] | None = None, prefix: Sequence[str] | None = None
    ):
        pass

    def subscribe(
        self, log_key: Sequence[str], cursor: str | None = None
    ) -> CapturedLogSubscription:
        return CapturedLogSubscription(self, log_key, cursor)

    def unsubscribe(self, subscription: CapturedLogSubscription):
        pass



from collections.abc import Callable, Mapping
from typing import Any, cast

from dagster_shared.record import copy, record
from typing_extensions import Self

import dagster._check as check
from dagster._core.definitions.decorators.schedule_decorator import schedule
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.metadata import RawMetadataMapping
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition.multi import MultiPartitionsDefinition
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.definition.static import StaticPartitionsDefinition
from dagster._core.definitions.partitions.definition.time_window import (
    TimeWindowPartitionsDefinition,
)
from dagster._core.definitions.partitions.utils.multi import (
    get_time_partitions_def,
    has_one_dimension_time_window_partitioning,
)
from dagster._core.definitions.run_request import RunRequest, SkipReason
from dagster._core.definitions.schedule_definition import (
    DefaultScheduleStatus,
    RunRequestIterator,
    ScheduleDefinition,
    ScheduleEvaluationContext,
)
from dagster._core.definitions.unresolved_asset_job_definition import UnresolvedAssetJobDefinition
from dagster._core.errors import DagsterInvalidDefinitionError


@record
class UnresolvedPartitionedAssetScheduleDefinition:
    """Points to an unresolved asset job. The asset selection isn't resolved yet, so we can't resolve
    the PartitionsDefinition, so we can't resolve the schedule cadence.
    """

    name: str
    job: UnresolvedAssetJobDefinition
    description: str | None
    default_status: DefaultScheduleStatus
    minute_of_hour: int | None
    hour_of_day: int | None
    day_of_week: int | None
    day_of_month: int | None
    tags: Mapping[str, str] | None
    metadata: Mapping[str, Any] | None

    def resolve(self, resolved_job: JobDefinition) -> ScheduleDefinition:
        partitions_def = resolved_job.partitions_def
        if partitions_def is None:
            check.failed(
                f"Job '{resolved_job.name}' provided to build_schedule_from_partitioned_job must"
                " contain partitioned assets or a partitions definition."
            )

        partitions_def = _check_valid_schedule_partitions_def(partitions_def)
        time_partitions_def = check.not_none(get_time_partitions_def(partitions_def))

        return schedule(
            name=self.name,
            cron_schedule=time_partitions_def.get_cron_schedule(
                self.minute_of_hour, self.hour_of_day, self.day_of_week, self.day_of_month
            ),
            job=resolved_job,
            default_status=self.default_status,
            execution_timezone=time_partitions_def.timezone,
            description=self.description,
            metadata=self.metadata,
        )(_get_schedule_evaluation_fn(partitions_def, resolved_job, self.tags))

    def with_metadata(self, metadata: RawMetadataMapping) -> Self:
        return copy(self, metadata=metadata)


def build_schedule_from_partitioned_job(
    job: JobDefinition | UnresolvedAssetJobDefinition,
    description: str | None = None,
    name: str | None = None,
    minute_of_hour: int | None = None,
    hour_of_day: int | None = None,
    day_of_week: int | None = None,
    day_of_month: int | None = None,
    default_status: DefaultScheduleStatus = DefaultScheduleStatus.STOPPED,
    tags: Mapping[str, str] | None = None,
    cron_schedule: str | None = None,
    execution_timezone: str | None = None,
    metadata: RawMetadataMapping | None = None,
) -> UnresolvedPartitionedAssetScheduleDefinition | ScheduleDefinition:
    """Creates a schedule from a job that targets
    time window-partitioned or statically-partitioned assets. The job can also be
    multi-partitioned, as long as one of the partition dimensions is time-partitioned.

    The schedule executes at the cadence specified by the time partitioning of the job or assets.

    **Example:**
        .. code-block:: python

            ######################################
            # Job that targets partitioned assets
            ######################################

            from dagster import (
                DailyPartitionsDefinition,
                asset,
                build_schedule_from_partitioned_job,
                define_asset_job,
                Definitions,
            )

            @asset(partitions_def=DailyPartitionsDefinition(start_date="2020-01-01"))
            def asset1():
                ...

            asset1_job = define_asset_job("asset1_job", selection=[asset1])

            # The created schedule will fire daily
            asset1_job_schedule = build_schedule_from_partitioned_job(asset1_job)

            Definitions(assets=[asset1], schedules=[asset1_job_schedule])

            ################
            # Non-asset job
            ################

            from dagster import DailyPartitionsDefinition, build_schedule_from_partitioned_job, jog


            @job(partitions_def=DailyPartitionsDefinition(start_date="2020-01-01"))
            def do_stuff_partitioned():
                ...

            # The created schedule will fire daily
            do_stuff_partitioned_schedule = build_schedule_from_partitioned_job(
                do_stuff_partitioned,
            )

            Definitions(schedules=[do_stuff_partitioned_schedule])
    """
    check.invariant(
        not (day_of_week and day_of_month),
        "Cannot provide both day_of_month and day_of_week parameter to"
        " build_schedule_from_partitioned_job.",
    )

    check.invariant(
        not (
            (cron_schedule or execution_timezone)
            and (
                day_of_month is not None
                or day_of_week is not None
                or hour_of_day is not None
                or minute_of_hour is not None
            )
        ),
        "Cannot provide both cron_schedule / execution_timezone parameters and"
        " day_of_month / day_of_week / hour_of_day / minute_of_hour parameters to"
        " build_schedule_from_partitioned_job.",
    )

    if isinstance(job, UnresolvedAssetJobDefinition) and job.partitions_def is None:
        if cron_schedule or execution_timezone:
            check.failed(
                "Cannot provide cron_schedule or execution_timezone to"
                " build_schedule_from_partitioned_job for a time-partitioned job."
            )

        return UnresolvedPartitionedAssetScheduleDefinition(
            job=job,
            default_status=default_status,
            name=check.opt_str_param(name, "name", f"{job.name}_schedule"),
            description=check.opt_str_param(description, "description"),
            minute_of_hour=minute_of_hour,
            hour_of_day=hour_of_day,
            day_of_week=day_of_week,
            day_of_month=day_of_month,
            tags=tags,
            metadata=metadata,
        )
    else:
        partitions_def = job.partitions_def
        if partitions_def is None:
            check.failed("The provided job is not partitioned")

        partitions_def = _check_valid_schedule_partitions_def(partitions_def)
        if isinstance(partitions_def, StaticPartitionsDefinition):
            check.not_none(
                cron_schedule,
                "Creating a schedule from a static partitions definition requires a cron schedule",
            )
            should_execute = None
        else:
            if cron_schedule or execution_timezone:
                check.failed(
                    "Cannot provide cron_schedule or execution_timezone to"
                    " build_schedule_from_partitioned_job for a time-partitioned job."
                )
            time_partitions_def = check.not_none(get_time_partitions_def(partitions_def))
            cron_schedule = time_partitions_def.get_cron_schedule(
                minute_of_hour, hour_of_day, day_of_week, day_of_month
            )
            execution_timezone = time_partitions_def.timezone
            if time_partitions_def.exclusions:

                def _should_execute(context: ScheduleEvaluationContext) -> bool:
                    with partition_loading_context(
                        effective_dt=context.scheduled_execution_time,
                        dynamic_partitions_store=context.instance
                        if context.instance_ref is not None
                        else None,
                    ):
                        window = time_partitions_def.get_last_partition_window_ignoring_exclusions()
                        if not window:
                            return True
                        return not time_partitions_def.is_window_start_excluded(window.start)

                should_execute = _should_execute
            else:
                should_execute = None

        return schedule(
            cron_schedule=cron_schedule,  # type: ignore[arg-type]
            job=job,
            default_status=default_status,
            execution_timezone=execution_timezone,
            name=check.opt_str_param(name, "name", f"{job.name}_schedule"),
            description=check.opt_str_param(description, "description"),
            should_execute=should_execute,
        )(_get_schedule_evaluation_fn(partitions_def, job, tags))


def _get_schedule_evaluation_fn(
    partitions_def: PartitionsDefinition,
    job: JobDefinition | UnresolvedAssetJobDefinition,
    tags: Mapping[str, str] | None = None,
) -> Callable[[ScheduleEvaluationContext], SkipReason | RunRequest | RunRequestIterator]:
    def schedule_fn(context):
        # Run for the latest partition. Prior partitions will have been handled by prior ticks.
        with partition_loading_context(
            effective_dt=context._scheduled_execution_time,  # noqa
            dynamic_partitions_store=context.instance if context.instance_ref is not None else None,
        ):
            if isinstance(partitions_def, TimeWindowPartitionsDefinition):
                partition_key = partitions_def.get_last_partition_key()
                if partition_key is None:
                    return SkipReason("The job's PartitionsDefinition has no partitions")

                return job.run_request_for_partition(
                    partition_key=partition_key, run_key=partition_key, tags=tags
                )
            if isinstance(partitions_def, StaticPartitionsDefinition):
                return [
                    job.run_request_for_partition(partition_key=key, run_key=key, tags=tags)
                    for key in partitions_def.get_partition_keys()
                ]
            else:
                check.invariant(isinstance(partitions_def, MultiPartitionsDefinition))
                time_window_dimension = partitions_def.time_window_dimension  # pyright: ignore[reportAttributeAccessIssue]
                partition_key = time_window_dimension.partitions_def.get_last_partition_key()
                if partition_key is None:
                    return SkipReason("The job's PartitionsDefinition has no partitions")

                return [
                    job.run_request_for_partition(partition_key=key, run_key=key, tags=tags)
                    for key in partitions_def.get_multipartition_keys_with_dimension_value(  # pyright: ignore[reportAttributeAccessIssue]
                        time_window_dimension.name, partition_key
                    )
                ]

    return schedule_fn  # pyright: ignore[reportReturnType]


def _check_valid_schedule_partitions_def(
    partitions_def: PartitionsDefinition,
) -> TimeWindowPartitionsDefinition | MultiPartitionsDefinition | StaticPartitionsDefinition:
    if not has_one_dimension_time_window_partitioning(partitions_def) and not isinstance(
        partitions_def, StaticPartitionsDefinition
    ):
        raise DagsterInvalidDefinitionError(
            "Tried to build a partitioned schedule from an asset job, but received an invalid"
            " partitions definition. The permitted partitions definitions are: \n1."
            " TimeWindowPartitionsDefinition\n2. MultiPartitionsDefinition with a single"
            " TimeWindowPartitionsDefinition dimension\n3. StaticPartitionsDefinition"
        )

    return cast(
        "TimeWindowPartitionsDefinition | MultiPartitionsDefinition | StaticPartitionsDefinition",
        partitions_def,
    )


schedule_from_partitions = build_schedule_from_partitioned_job



import logging
import uuid
from collections import defaultdict
from collections.abc import Callable
from contextlib import contextmanager
from typing import Any

import sqlalchemy as db
from sqlalchemy.pool import NullPool

from dagster._core.storage.event_log.base import EventLogCursor
from dagster._core.storage.event_log.schema import SqlEventLogStorageMetadata
from dagster._core.storage.event_log.sql_event_log import SqlEventLogStorage
from dagster._core.storage.sql import create_engine, get_alembic_config, stamp_alembic_rev
from dagster._core.storage.sqlite import create_in_memory_conn_string
from dagster._serdes import ConfigurableClass
from dagster._serdes.config_class import ConfigurableClassData


class InMemoryEventLogStorage(SqlEventLogStorage, ConfigurableClass):
    """In memory only event log storage. Used by ephemeral DagsterInstance or for testing purposes.

    WARNING: The Dagster UI and other core functionality will not work if this is used on a real DagsterInstance
    """

    def __init__(self, inst_data: ConfigurableClassData | None = None, preload=None):
        self._inst_data = inst_data
        self._engine = create_engine(
            create_in_memory_conn_string(f"events-{uuid.uuid4()}"),
            poolclass=NullPool,
        )
        self._handlers = defaultdict(set)
        self._storage_id = 0  # mirror the storage id, to mimic watching cursors

        # hold one connection for life of instance, but vend new ones for specific calls
        self._held_conn = self._engine.connect()
        with self._held_conn.begin():
            SqlEventLogStorageMetadata.create_all(self._held_conn)
            alembic_config = get_alembic_config(__file__, "sqlite/alembic/alembic.ini")
            stamp_alembic_rev(alembic_config, self._held_conn)

        if preload:
            for payload in preload:
                for event in payload.event_list:
                    self.store_event(event)

    def has_secondary_index(self, name: str) -> bool:
        return True

    @contextmanager
    def _connect(self):
        with self._engine.connect() as conn:
            with conn.begin():
                conn.execute(db.text("PRAGMA journal_mode=WAL;"))
                conn.execute(db.text("PRAGMA foreign_keys=ON;"))
                yield conn

    def run_connection(self, run_id=None):
        return self._connect()

    def index_connection(self):
        return self._connect()

    def has_table(self, table_name: str) -> bool:
        with self._engine.connect() as conn:
            return bool(self._engine.dialect.has_table(conn, table_name))

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @classmethod
    def from_config_value(cls, inst_data, config_value):
        return cls(inst_data)

    def upgrade(self):
        pass

    def store_event(self, event):
        super().store_event(event)
        self._storage_id += 1

        handlers = list(self._handlers[event.run_id])
        for handler in handlers:
            try:
                handler(event, str(EventLogCursor.from_storage_id(self._storage_id)))
            except Exception:
                logging.exception("Exception in callback for event watch on run %s.", event.run_id)

    def watch(self, run_id: str, cursor: str, callback: Callable[..., Any]):  # pyright: ignore[reportIncompatibleMethodOverride]
        self._handlers[run_id].add(callback)

    def end_watch(self, run_id: str, handler: Callable[..., Any]):
        if handler in self._handlers[run_id]:
            self._handlers[run_id].remove(handler)

    @property
    def is_persistent(self) -> bool:
        return False

    @property
    def supports_global_concurrency_limits(self) -> bool:  # pyright: ignore[reportIncompatibleVariableOverride]
        return False

    def dispose(self):
        self._held_conn.close()
        self._engine.dispose()



import logging
import os
import sys

import click
from dagster_shared.cli import workspace_options
from dagster_shared.error import remove_system_frames_from_error

from dagster import __version__ as dagster_version
from dagster._annotations import superseded
from dagster._cli.utils import assert_no_remaining_opts, get_possibly_temporary_instance_for_cli
from dagster._cli.workspace.cli_target import WorkspaceOpts, get_workspace_from_cli_opts
from dagster._utils.error import serializable_error_info_from_exc_info, unwrap_user_code_error
from dagster._utils.log import configure_loggers


@click.group(name="definitions")
def definitions_cli():
    """Commands for working with Dagster definitions."""


@workspace_options
@click.option(
    "--log-level",
    help="Set the log level for dagster services.",
    show_default=True,
    default="info",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
@click.option(
    "--log-format",
    type=click.Choice(["colored", "json", "rich"], case_sensitive=False),
    show_default=True,
    required=False,
    default="colored",
    help="Format of the logs for dagster services",
)
@click.option(
    "--load-with-grpc",
    flag_value=True,
    default=False,
    help="Load the code locations using a gRPC server, instead of in-process.",
)
@click.option(
    "--verbose",
    "-v",
    flag_value=True,
    default=False,
    help="Show verbose stack traces, including system frames in stack traces.",
)
@definitions_cli.command(
    name="validate",
    help="""
    The `dagster definitions validate` command loads and validate your Dagster definitions using a Dagster instance.

    This command indicates which code locations contain errors, and which ones can be successfully loaded.
    Code locations containing errors are considered invalid, otherwise valid.

    When running, this command sets the environment variable `DAGSTER_IS_DEFS_VALIDATION_CLI=1`.
    This environment variable can be used to control the behavior of your code in validation mode.

    This command returns an exit code 1 when errors are found, otherwise an exit code 0.

    This command should be run in a Python environment where the `dagster` package is installed.
    """,
)
@superseded(
    additional_warn_text="Use 'dg check defs' instead.",
    emit_runtime_warning=True,
)
def definitions_validate_command(
    log_level: str,
    log_format: str,
    load_with_grpc: bool,
    verbose: bool,
    **other_opts: object,
):
    definitions_validate_command_impl(
        log_level=log_level,
        log_format=log_format,
        allow_in_process=not load_with_grpc,
        verbose=verbose,
        **other_opts,
    )


def definitions_validate_command_impl(
    log_level: str,
    log_format: str,
    allow_in_process: bool,
    verbose: bool,
    **other_opts: object,
):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    os.environ["DAGSTER_IS_DEFS_VALIDATION_CLI"] = "1"

    configure_loggers(formatter=log_format, log_level=log_level.upper())
    logger = logging.getLogger("dagster")
    logging.captureWarnings(True)

    removed_system_frame_hint = lambda is_first_hidden_frame, i: (
        f"  [{i} dagster system frames hidden, run with --verbose to see the full stack trace]\n"
        if is_first_hidden_frame
        else f"  [{i} dagster system frames hidden]\n"
    )

    with get_possibly_temporary_instance_for_cli(
        "dagster definitions validate", logger=logger
    ) as instance:
        with get_workspace_from_cli_opts(
            instance=instance,
            version=dagster_version,
            workspace_opts=workspace_opts,
            allow_in_process=allow_in_process,
            log_level=log_level,
        ) as workspace:
            invalid_locations = [
                entry
                for entry in workspace.get_code_location_entries().values()
                if entry.load_error
            ]
            for code_location, entry in workspace.get_code_location_entries().items():
                if entry.load_error:
                    if verbose:
                        underlying_error = entry.load_error
                    else:
                        underlying_error = remove_system_frames_from_error(
                            unwrap_user_code_error(entry.load_error),
                            build_system_frame_removed_hint=removed_system_frame_hint,
                        )
                    logger.error(
                        f"Validation failed for code location {code_location}:\n\n"
                        f"{underlying_error.to_string()}"
                    )
                    pass
                else:
                    logger.info(f"Validation successful for code location {code_location}.")

            try:
                workspace.asset_graph.validate_partitions()
            except Exception:
                logger.error(
                    f"Asset graph contained an invalid partition mapping:\n{serializable_error_info_from_exc_info(sys.exc_info()).to_string()}"
                )
                sys.exit(1)

    if invalid_locations:
        logger.error(f"Validation for {len(invalid_locations)} code locations failed.")
        sys.exit(1)
    else:
        logger.info("All code locations passed validation.")



from collections.abc import Sequence
from enum import Enum

INCREASE_TIMEOUT_DAGSTER_YAML_MSG = """To increase the timeout, add the following to a dagster.yaml file, located in your $DAGSTER_HOME folder or the folder where you are running `dagster dev`:

code_servers:
  local_startup_timeout: <timeout value>

"""


class GrpcServerCommand(Enum):
    API_GRPC = "api_grpc"
    CODE_SERVER_START = "code_server_start"

    @property
    def server_command(self) -> Sequence[str]:
        return (
            ["api", "grpc", "--lazy-load-user-code"]
            if self == GrpcServerCommand.API_GRPC
            else ["code-server", "start"]
        )



from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from typing import Any

from dagster._core.definitions.assets.definition.asset_spec import (
    SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET,
    AssetSpec,
)
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata.metadata_set import TableMetadataSet
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._utils.warnings import beta_warning


def resolve_similar_asset_names(
    target_asset_key: AssetKey,
    asset_keys: Iterable[AssetKey],
) -> Sequence[AssetKey]:
    """Given a target asset key (an upstream dependency which we can't find), produces a list of
    similar asset keys from the list of asset definitions. We use this list to produce a helpful
    error message that can help users debug their asset dependencies.
    """
    similar_names: list[AssetKey] = []

    target_asset_key_split = ("/".join(target_asset_key.path)).split("/")

    for asset_key in asset_keys:
        *target_asset_key_prefix, target_asset_key_name = target_asset_key.path
        *asset_key_prefix, asset_key_name = asset_key.path

        try:
            from rapidfuzz import fuzz

            is_similar_name = bool(
                fuzz.ratio(asset_key_name, target_asset_key_name, score_cutoff=80)
            )
            is_similar_prefix = bool(
                fuzz.ratio(
                    " ".join(asset_key_prefix),
                    " ".join(target_asset_key_prefix),
                    score_cutoff=80,
                )
            )
        except ImportError:
            from difflib import get_close_matches

            is_similar_name = bool(
                get_close_matches(asset_key_name, [target_asset_key_name], cutoff=0.8)
            )
            is_similar_prefix = bool(
                get_close_matches(
                    " ".join(asset_key_prefix), [" ".join(target_asset_key_prefix)], cutoff=0.8
                )
            )

        # Whether the asset key or upstream key has the same prefix and a similar
        # name
        # e.g. [snowflake, elementl, key] and [snowflake, elementl, ey]
        is_same_prefix_similar_name = (
            asset_key_prefix == target_asset_key_prefix and is_similar_name
        )

        # Whether the asset key or upstream key has a similar prefix and the same
        # name
        # e.g. [snowflake, elementl, key] and [nowflake, elementl, key]
        is_similar_prefix_same_name = asset_key_name == target_asset_key_name and is_similar_prefix

        # Whether the asset key or upstream key has one more prefix component than
        # the other, and the same name
        # e.g. [snowflake, elementl, key] and [snowflake, elementl, prod, key]
        is_off_by_one_prefix_component_same_name = (
            asset_key.path[-1] == target_asset_key.path[-1]
            and len(set(asset_key.path).symmetric_difference(set(target_asset_key.path))) == 1
            and max(len(asset_key.path), len(target_asset_key.path)) > 1
        )

        # If the asset key provided has no prefix and the upstream key has
        # the same name but a prefix of any length
        no_prefix_but_is_match_with_prefix = (
            len(target_asset_key.path) == 1 and asset_key.path[-1] == target_asset_key.path[-1]
        )

        matches_slashes_turned_to_prefix_gaps = asset_key.path == target_asset_key_split

        if (
            is_same_prefix_similar_name
            or is_similar_prefix_same_name
            or is_off_by_one_prefix_component_same_name
            or no_prefix_but_is_match_with_prefix
            or matches_slashes_turned_to_prefix_gaps
        ):
            similar_names.append(asset_key)
    return sorted(similar_names, key=lambda key: key.to_string())


def _build_input_error_msg(
    input_name: str, upstream_key: AssetKey, asset_keys: Iterable[AssetKey]
) -> str:
    msg = f'Input asset "{upstream_key.to_string()}" is not produced by any of the provided asset ops and is not one of the provided sources.'
    similar_names = resolve_similar_asset_names(upstream_key, asset_keys)
    if similar_names:
        # Arbitrarily limit to 10 similar names to avoid a huge error message
        subset_similar_names = similar_names[:10]
        similar_to_string = ", ".join(similar.to_string() for similar in subset_similar_names)
        msg += f" Did you mean one of the following?\n\t{similar_to_string}"
    return msg


def resolve_assets_def_deps(assets_defs: list[AssetsDefinition]) -> list[AssetsDefinition]:
    specs = [spec for assets_def in assets_defs for spec in assets_def.specs]
    asset_keys = {spec.key for spec in specs}

    # build mapping from group and name to asset keys
    keys_by_group_and_name: dict[tuple[str | None, str], list[AssetKey]] = defaultdict(list)
    for spec in specs:
        if spec.group_name is not None:
            keys_by_group_and_name[(spec.group_name, spec.key.path[-1])].append(spec.key)

    # build mapping from table name to asset keys
    keys_by_table_name: dict[str | None, list[AssetKey]] = defaultdict(list)
    for spec in specs:
        table_name = TableMetadataSet.extract(spec.metadata).table_name
        if table_name is not None:
            keys_by_table_name[table_name].append(spec.key)

    # analyze each assets definition for unresolved dependencies and attempt to resolve them
    warned = False
    resolved_assets_defs: list[AssetsDefinition] = []
    for assets_def in assets_defs:
        replacements: dict[AssetKey, AssetKey] = {}
        for spec in assets_def.specs:
            for dep in spec.deps:
                # dep's key already maps to an existing spec
                if dep.asset_key in asset_keys:
                    continue

                input_name = assets_def.input_names_by_node_key.get(dep.asset_key)
                input_def = assets_def.node_def.input_def_named(input_name) if input_name else None
                table_name = TableMetadataSet.extract(dep.metadata).table_name

                # attempt to match by name and group
                matching_asset_keys = keys_by_group_and_name[
                    (spec.group_name, dep.asset_key.path[0])
                ]
                if (
                    # no prefix explicitly provided
                    len(dep.asset_key.path) == 1
                    # unambiguous match
                    and len(matching_asset_keys) == 1
                    # not remapping to self
                    and matching_asset_keys[0] not in assets_def.keys
                ):
                    replacements[dep.asset_key] = matching_asset_keys[0]
                    if not warned:
                        beta_warning(
                            f"Asset {spec.key.to_string()}'s dependency"
                            f" '{input_name or dep.asset_key.to_string()}' was resolved to upstream asset"
                            f" {matching_asset_keys[0].to_string()}, because the name matches and they're in the"
                            " same group. This is a beta functionality that may change in a"
                            " future release"
                        )
                        warned = True
                    continue

                # attempt to match by table name
                matching_asset_keys = keys_by_table_name[table_name]
                if len(matching_asset_keys) > 0:
                    if len(matching_asset_keys) > 1:
                        raise DagsterInvalidDefinitionError(
                            f'Multiple upstream assets found with table name "{table_name}": {matching_asset_keys}. '
                            f"Unable to resolve dependency {dep.asset_key.to_string()} for asset {spec.key.to_string()}."
                        )
                    replacements[dep.asset_key] = matching_asset_keys[0]
                    if not warned:
                        beta_warning(
                            f"Asset {spec.key.to_string()}'s dependency"
                            f" '{input_name or dep.asset_key.to_string()}' was resolved to upstream asset"
                            f" {matching_asset_keys[0].to_string()}, because the table name matches. This is a beta functionality that may change in a"
                            " future release"
                        )
                        warned = True
                    continue

                # could not resolve the dependency, and input requires a real upstream assets def
                if input_name and input_def and not input_def.dagster_type.is_nothing:
                    raise DagsterInvalidDefinitionError(
                        _build_input_error_msg(input_name, dep.asset_key, asset_keys)
                    )

        # only update the object if replacements were necessary
        resolved_assets_defs.append(
            assets_def.with_attributes(asset_key_replacements=replacements)
            if replacements
            else assets_def
        )

    return resolved_assets_defs


def resolve_stub_assets_defs(assets_defs: list[AssetsDefinition]) -> list[AssetsDefinition]:
    defined_keys: set[AssetKey] = set()
    deps_metadata_by_key: dict[AssetKey, list[Mapping[str, Any]]] = defaultdict(list)

    for assets_def in assets_defs:
        for spec in assets_def.specs:
            defined_keys.add(spec.key)
            for dep in spec.deps:
                deps_metadata_by_key[dep.asset_key].append(dep.metadata)
        # treat the asset keys defined for asset checks as AssetDeps with no metadata
        for check_spec in assets_def.check_specs:
            deps_metadata_by_key[check_spec.asset_key].append({})

    stub_assets_defs: list[AssetsDefinition] = []

    # iterate through the set of keys that are a dependency but have no definition
    for key in deps_metadata_by_key.keys() - defined_keys:
        # combine metadata from all dependencies and error if there are conflicts
        metadata = {}
        for dep_metadata in deps_metadata_by_key[key]:
            if not dep_metadata:
                continue

            # metadata set, but conflicts
            if metadata and dep_metadata != metadata:
                raise DagsterInvalidDefinitionError(
                    f"Conflicting metadata found on AssetDeps referencing {key.to_string()}: {metadata} != {dep_metadata}. "
                    "All metadata for AssetDeps referencing a given key must be identical or empty."
                )
            metadata = dict(dep_metadata)

        # add in flag to indicate that this is a stub asset
        metadata[SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET] = True
        stub_assets_defs.append(AssetsDefinition(specs=[AssetSpec(key=key, metadata=metadata)]))

    return stub_assets_defs



from collections.abc import Mapping, Sequence
from datetime import datetime  # noqa
from os import PathLike  # noqa
from typing import Any, Generic, TypeVar

from dagster_shared.check.decorator import checked
from dagster_shared.record import IHaveNew, LegacyNamedTupleMixin, record_custom

import dagster._check as check
from dagster._annotations import PublicAttr, public
from dagster._core.definitions.asset_checks.asset_check_result import AssetCheckResult
from dagster._core.definitions.data_version import DataVersion
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.metadata import (  # noqa
    MetadataValue,
    RawMetadataMapping,
    TableColumnLineage,
    TableSchema,
)
from dagster._core.definitions.utils import NoValueSentinel


@checked
def coerce_asset_result_args(
    asset_key: CoercibleToAssetKey | None,
    metadata: RawMetadataMapping | None,
    check_results: Sequence[AssetCheckResult] | None,
    data_version: DataVersion | None,
    tags: Mapping[str, Any] | None,
) -> Mapping[str, Any]:
    from dagster._core.definitions.events import validate_asset_event_tags

    return dict(
        asset_key=AssetKey.from_coercible(asset_key) if asset_key else None,
        metadata=metadata,
        check_results=check_results or [],
        data_version=data_version,
        tags=validate_asset_event_tags(tags),
    )


class AssetResult:
    """Base class for MaterializeResult and ObserveResult."""

    asset_key: PublicAttr[AssetKey | None]
    metadata: PublicAttr[RawMetadataMapping | None]
    check_results: PublicAttr[Sequence[AssetCheckResult]]
    data_version: PublicAttr[DataVersion | None]
    tags: PublicAttr[Mapping[str, str] | None]

    def check_result_named(self, check_name: str) -> AssetCheckResult:
        for check_result in self.check_results:
            if check_result.check_name == check_name:
                return check_result

        check.failed(f"Could not find check result named {check_name}")


T = TypeVar("T")


@record_custom(checked=False)
@public
class MaterializeResult(AssetResult, Generic[T], IHaveNew, LegacyNamedTupleMixin):
    """An object representing a successful materialization of an asset. These can be returned from
    @asset and @multi_asset decorated functions to pass metadata or specify specific assets were
    materialized.

    Args:
        asset_key (Optional[AssetKey]): Optional in @asset, required in @multi_asset to discern which asset this refers to.
        metadata (Optional[RawMetadataMapping]): Metadata to record with the corresponding AssetMaterialization event.
        check_results (Optional[Sequence[AssetCheckResult]]): Check results to record with the
            corresponding AssetMaterialization event.
        data_version (Optional[DataVersion]): The data version of the asset that was observed.
        tags (Optional[Mapping[str, str]]): Tags to record with the corresponding
            AssetMaterialization event.
        value (Optional[Any]): The output value of the asset that was materialized.
    """

    asset_key: PublicAttr[AssetKey | None]
    metadata: PublicAttr[RawMetadataMapping | None]
    check_results: PublicAttr[Sequence[AssetCheckResult]]
    data_version: PublicAttr[DataVersion | None]
    tags: PublicAttr[Mapping[str, str] | None]
    value: PublicAttr[T]

    def __new__(
        cls,
        asset_key: CoercibleToAssetKey | None = None,
        metadata: RawMetadataMapping | None = None,
        check_results: Sequence[AssetCheckResult] | None = None,
        data_version: DataVersion | None = None,
        tags: Mapping[str, str] | None = None,
        value: T = NoValueSentinel,
    ):
        return super().__new__(
            cls,
            **coerce_asset_result_args(
                asset_key=asset_key,
                metadata=metadata,
                check_results=check_results,
                data_version=data_version,
                tags=tags,
            ),
            value=value,
        )


@public
@record_custom(checked=False)
class ObserveResult(AssetResult, IHaveNew, LegacyNamedTupleMixin):
    """An object representing a successful observation of an asset. These can be returned from an
    @observable_source_asset decorated function to pass metadata.

    Args:
        asset_key (Optional[AssetKey]): The asset key. Optional to include.
        metadata (Optional[RawMetadataMapping]): Metadata to record with the corresponding
            AssetObservation event.
        check_results (Optional[Sequence[AssetCheckResult]]): Check results to record with the
            corresponding AssetObservation event.
        data_version (Optional[DataVersion]): The data version of the asset that was observed.
        tags (Optional[Mapping[str, str]]): Tags to record with the corresponding AssetObservation
            event.
    """

    asset_key: PublicAttr[AssetKey | None]
    metadata: PublicAttr[RawMetadataMapping | None]
    check_results: PublicAttr[Sequence[AssetCheckResult]]
    data_version: PublicAttr[DataVersion | None]
    tags: PublicAttr[Mapping[str, str] | None]

    def __new__(
        cls,
        asset_key: CoercibleToAssetKey | None = None,
        metadata: RawMetadataMapping | None = None,
        check_results: Sequence[AssetCheckResult] | None = None,
        data_version: DataVersion | None = None,
        tags: Mapping[str, str] | None = None,
    ):
        return super().__new__(
            cls,
            **coerce_asset_result_args(
                asset_key=asset_key,
                metadata=metadata,
                check_results=check_results,
                data_version=data_version,
                tags=tags,
            ),
        )



# Re-export all public symbols for backward compatibility
from dagster._symbol_annotations.annotatable import Annotatable, T_Annotatable, copy_annotations
from dagster._symbol_annotations.lifecycle import (
    BetaInfo,
    DeprecatedInfo,
    PreviewInfo,
    SupersededInfo,
    attach_deprecation_info_and_wrap,
    beta,
    beta_param,
    deprecated,
    deprecated_param,
    get_beta_info,
    get_beta_param_info,
    get_beta_params,
    get_deprecated_info,
    get_deprecated_param_info,
    get_deprecated_params,
    get_preview_info,
    get_superseded_info,
    has_beta_params,
    has_deprecated_params,
    hidden_param,
    is_beta,
    is_beta_param,
    is_deprecated,
    is_deprecated_param,
    is_preview,
    is_superseded,
    only_allow_hidden_params_in_kwargs,
    preview,
    superseded,
)
from dagster._symbol_annotations.public import PUBLIC, PublicAttr, is_public, public

__all__ = [
    "PUBLIC",
    "Annotatable",
    "BetaInfo",
    "DeprecatedInfo",
    "PreviewInfo",
    "PublicAttr",
    "SupersededInfo",
    "T_Annotatable",
    "attach_deprecation_info_and_wrap",
    "beta",
    "beta_param",
    "copy_annotations",
    "deprecated",
    "deprecated_param",
    "get_beta_info",
    "get_beta_param_info",
    "get_beta_params",
    "get_deprecated_info",
    "get_deprecated_param_info",
    "get_deprecated_params",
    "get_preview_info",
    "get_superseded_info",
    "has_beta_params",
    "has_deprecated_params",
    "hidden_param",
    "is_beta",
    "is_beta_param",
    "is_deprecated",
    "is_deprecated_param",
    "is_preview",
    "is_public",
    "is_superseded",
    "only_allow_hidden_params_in_kwargs",
    "preview",
    "public",
    "superseded",
]



"""This module contains data objects meant to be serialized between
host processes and user processes. They should contain no
business logic or clever indexing. Use the classes in external.py
for that.
"""

from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Annotated

from dagster_shared.record import ImportFrom
from typing_extensions import Self

from dagster import _check as check
from dagster._core.definitions.partitions.schedule_type import ScheduleType
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._record import IHaveNew, record, record_custom
from dagster._serdes import whitelist_for_serdes
from dagster._time import datetime_from_timestamp

if TYPE_CHECKING:
    from dagster._core.definitions.partitions.definition import (
        DynamicPartitionsDefinition,
        MultiPartitionsDefinition,
        PartitionsDefinition,
        StaticPartitionsDefinition,
        TimeWindowPartitionsDefinition,
    )
    from dagster._core.definitions.timestamp import TimestampWithTimezone


class PartitionsSnap(ABC):
    @classmethod
    def from_def(cls, partitions_def: "PartitionsDefinition") -> "PartitionsSnap":
        from dagster._core.definitions.partitions.definition import (
            DynamicPartitionsDefinition,
            MultiPartitionsDefinition,
            StaticPartitionsDefinition,
            TimeWindowPartitionsDefinition,
        )

        if isinstance(partitions_def, TimeWindowPartitionsDefinition):
            return TimeWindowPartitionsSnap.from_def(partitions_def)
        elif isinstance(partitions_def, StaticPartitionsDefinition):
            return StaticPartitionsSnap.from_def(partitions_def)
        elif isinstance(partitions_def, MultiPartitionsDefinition):
            return MultiPartitionsSnap.from_def(partitions_def)
        elif isinstance(partitions_def, DynamicPartitionsDefinition):
            return DynamicPartitionsSnap.from_def(partitions_def)
        else:
            raise DagsterInvalidDefinitionError(
                "Only static, time window, multi-dimensional partitions, and dynamic partitions"
                " definitions with a name parameter are currently supported."
            )

    @abstractmethod
    def get_partitions_definition(self) -> "PartitionsDefinition": ...


@whitelist_for_serdes(storage_name="ExternalTimeWindowPartitionsDefinitionData")
@record
class TimeWindowPartitionsSnap(PartitionsSnap):
    start: float
    timezone: str | None
    fmt: str
    end_offset: int
    end: float | None = None
    cron_schedule: str | None = None
    exclusions: (
        Sequence[
            str
            | Annotated["TimestampWithTimezone", ImportFrom("dagster._core.definitions.timestamp")]
        ]
        | None
    ) = None
    # superseded by cron_schedule, but kept around for backcompat
    schedule_type: ScheduleType | None = None
    # superseded by cron_schedule, but kept around for backcompat
    minute_offset: int | None = None
    # superseded by cron_schedule, but kept around for backcompat
    hour_offset: int | None = None
    # superseded by cron_schedule, but kept around for backcompat
    day_offset: int | None = None

    @classmethod
    def from_def(cls, partitions_def: "TimeWindowPartitionsDefinition") -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        from dagster._core.definitions.partitions.definition import TimeWindowPartitionsDefinition

        check.inst_param(partitions_def, "partitions_def", TimeWindowPartitionsDefinition)
        return cls(
            cron_schedule=partitions_def.cron_schedule,
            start=partitions_def.start.timestamp(),
            end=partitions_def.end.timestamp() if partitions_def.end else None,
            timezone=partitions_def.timezone,
            fmt=partitions_def.fmt,
            end_offset=partitions_def.end_offset,
            exclusions=partitions_def.exclusions if partitions_def.exclusions else None,
        )

    def get_partitions_definition(self):
        from dagster._core.definitions.partitions.definition import TimeWindowPartitionsDefinition

        if self.cron_schedule is not None:
            return TimeWindowPartitionsDefinition(
                cron_schedule=self.cron_schedule,
                start=datetime_from_timestamp(self.start, tz=self.timezone),  # pyright: ignore[reportArgumentType]
                timezone=self.timezone,
                fmt=self.fmt,
                end_offset=self.end_offset,
                end=(datetime_from_timestamp(self.end, tz=self.timezone) if self.end else None),  # pyright: ignore[reportArgumentType]
                exclusions=self.exclusions if self.exclusions else None,
            )
        else:
            # backcompat case
            return TimeWindowPartitionsDefinition(
                schedule_type=self.schedule_type,
                start=datetime_from_timestamp(self.start, tz=self.timezone),  # pyright: ignore[reportArgumentType]
                timezone=self.timezone,
                fmt=self.fmt,
                end_offset=self.end_offset,
                end=(datetime_from_timestamp(self.end, tz=self.timezone) if self.end else None),  # pyright: ignore[reportArgumentType]
                minute_offset=self.minute_offset,
                hour_offset=self.hour_offset,
                day_offset=self.day_offset,
                exclusions=self.exclusions if self.exclusions else None,
            )


def _dedup_partition_keys(keys: Sequence[str]) -> Sequence[str]:
    # Use both a set and a list here to preserve lookup performance in case of large inputs. (We
    # can't just use a set because we need to preserve ordering.)
    seen_keys: set[str] = set()
    new_keys: list[str] = []
    for key in keys:
        if key not in seen_keys:
            new_keys.append(key)
            seen_keys.add(key)
    return new_keys


@whitelist_for_serdes(storage_name="ExternalStaticPartitionsDefinitionData")
@record_custom(checked=False)
class StaticPartitionsSnap(PartitionsSnap, IHaveNew):
    partition_keys: Sequence[str]

    def __new__(cls, partition_keys: Sequence[str]):
        # for back compat reasons we allow str as a Sequence[str] here
        if not isinstance(partition_keys, str):
            check.sequence_param(
                partition_keys,
                "partition_keys",
                of_type=str,
            )

        return super().__new__(
            cls,
            partition_keys=partition_keys,
        )

    @classmethod
    def from_def(cls, partitions_def: "StaticPartitionsDefinition") -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        from dagster._core.definitions.partitions.definition import StaticPartitionsDefinition

        check.inst_param(partitions_def, "partitions_def", StaticPartitionsDefinition)
        return cls(partition_keys=partitions_def.get_partition_keys())

    def get_partitions_definition(self):
        from dagster._core.definitions.partitions.definition import StaticPartitionsDefinition

        # v1.4 made `StaticPartitionsDefinition` error if given duplicate keys. This caused
        # host process errors for users who had not upgraded their user code to 1.4 and had dup
        # keys, since the host process `StaticPartitionsDefinition` would throw an error.
        keys = _dedup_partition_keys(self.partition_keys)
        return StaticPartitionsDefinition(keys)


@whitelist_for_serdes(
    storage_name="ExternalPartitionDimensionDefinition",
    storage_field_names={"partitions": "external_partitions_def_data"},
)
@record
class PartitionDimensionSnap:
    name: str
    partitions: PartitionsSnap


@whitelist_for_serdes(
    storage_name="ExternalMultiPartitionsDefinitionData",
    storage_field_names={"partition_dimensions": "external_partition_dimension_definitions"},
)
@record
class MultiPartitionsSnap(PartitionsSnap):
    partition_dimensions: Sequence[PartitionDimensionSnap]

    @classmethod
    def from_def(cls, partitions_def: "MultiPartitionsDefinition") -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        from dagster._core.definitions.partitions.definition import MultiPartitionsDefinition

        check.inst_param(partitions_def, "partitions_def", MultiPartitionsDefinition)

        return cls(
            partition_dimensions=[
                PartitionDimensionSnap(
                    name=dimension.name,
                    partitions=PartitionsSnap.from_def(dimension.partitions_def),
                )
                for dimension in partitions_def.partitions_defs
            ]
        )

    def get_partitions_definition(self):
        from dagster._core.definitions.partitions.definition import MultiPartitionsDefinition

        return MultiPartitionsDefinition(
            {
                partition_dimension.name: (
                    partition_dimension.partitions.get_partitions_definition()
                )
                for partition_dimension in self.partition_dimensions
            }
        )


@whitelist_for_serdes(storage_name="ExternalDynamicPartitionsDefinitionData")
@record
class DynamicPartitionsSnap(PartitionsSnap):
    name: str

    @classmethod
    def from_def(cls, partitions_def: "DynamicPartitionsDefinition") -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        from dagster._core.definitions.partitions.definition import DynamicPartitionsDefinition

        check.inst_param(partitions_def, "partitions_def", DynamicPartitionsDefinition)
        if partitions_def.name is None:
            raise DagsterInvalidDefinitionError(
                "Dagster does not support dynamic partitions definitions without a name parameter."
            )
        return cls(name=partitions_def.name)

    def get_partitions_definition(self):
        from dagster._core.definitions.partitions.definition import DynamicPartitionsDefinition

        return DynamicPartitionsDefinition(name=self.name)



import re


def clean_name(name: str) -> str:
    """Cleans an input to be a valid Dagster name by replacing invalid characters with underscores."""
    return re.sub(r"[^A-Za-z0-9_]+", "_", name)


def clean_name_lower(name: str) -> str:
    """Cleans an input to be a valid Dagster name by replacing invalid characters with underscores
    and converting uppercase characters into lowercase characters.
    """
    return clean_name(name=name).lower()


def clean_name_lower_with_dots(name: str) -> str:
    """Cleans an input to be a valid Dagster name by replacing invalid characters with underscores
    and converting uppercase characters into lowercase characters. This function keeps dots from the original string.
    """
    return re.sub(r"[^A-Za-z0-9_.]+", "_", name).lower()



import importlib
from pathlib import Path
from types import ModuleType

import dagster_shared.check as check
from dagster_shared.serdes.objects.package_entry import json_for_all_components
from dagster_shared.utils.warnings import normalize_renamed_param

from dagster._annotations import deprecated, public
from dagster._core.definitions.definitions_class import Definitions
from dagster._symbol_annotations.lifecycle import deprecated_param
from dagster._utils.warnings import suppress_dagster_warnings
from dagster.components.component.component import Component
from dagster.components.core.component_tree import ComponentTree, LegacyAutoloadingComponentTree

PLUGIN_COMPONENT_TYPES_JSON_METADATA_KEY = "plugin_component_types_json"


@deprecated(breaking_version="0.2.0")
@suppress_dagster_warnings
def build_component_defs(components_root: Path) -> Definitions:
    """Build a Definitions object for all the component instances in a given project.

    Args:
        components_root (Path): The path to the components root. This is a directory containing
            subdirectories with component instances.
    """
    defs_root = importlib.import_module(
        f"{Path(components_root).parent.name}.{Path(components_root).name}"
    )

    return load_defs(defs_root=defs_root, project_root=components_root.parent.parent)


def get_project_root(defs_root: ModuleType) -> Path:
    """Find the project root directory containing pyproject.toml or setup.py.

    Args:
        defs_root: A module object from which to start the search.

    Returns:
        The absolute path to the project root directory.

    Raises:
        FileNotFoundError: If no project root with pyproject.toml or setup.py is found.
    """
    # Get the module's file path
    module_path = getattr(defs_root, "__file__", None)
    if module_path is None:
        # For modules without __file__ attribute (e.g. namespace packages), try to get path from __path__
        module_paths = getattr(defs_root, "__path__", None)
        if module_paths and len(module_paths) > 0:
            module_path = module_paths[0]
    if not module_path:
        raise FileNotFoundError(f"Module {defs_root} has no __file__ attribute")

    # Start with the directory containing the module
    current_dir = Path(module_path).parent

    # Traverse up until we find pyproject.toml or setup.py
    while current_dir != current_dir.parent:  # Stop at root
        if (current_dir / "pyproject.toml").exists() or (current_dir / "setup.py").exists():
            return current_dir
        current_dir = current_dir.parent

    raise FileNotFoundError("No project root with pyproject.toml or setup.py found")


@public
@suppress_dagster_warnings
def build_defs_for_component(component: Component) -> Definitions:
    """Constructs Definitions from a standalone component. This is useful for
    loading individual components in a non-component project.

    Args:
        component (Component): The component to load defs from.
    """
    return component.build_defs(ComponentTree.for_test().load_context)


@public
@deprecated_param(
    param="project_root",
    breaking_version="2.0",
    additional_warn_text="Use `path_within_project` instead.",
)
@suppress_dagster_warnings
def load_from_defs_folder(
    *,
    path_within_project: Path | None = None,
    project_root: Path | None = None,
) -> Definitions:
    """Constructs a Definitions object by automatically discovering and loading all Dagster
    definitions from a project's defs folder structure.

    This function serves as the primary entry point for loading definitions in dg-managed
    projects. It reads the project configuration (dg.toml or pyproject.toml), identifies
    the defs module, and recursively loads all components, assets, jobs, and other Dagster
    definitions from the project structure.

    The function automatically handles:

    * Reading project configuration to determine the defs module location
    * Importing and traversing the defs module hierarchy
    * Loading component definitions and merging them into a unified Definitions object
    * Enriching definitions with plugin component metadata from entry points

    Args:
        path_within_project (Path): A path within the dg project directory.
            This directory or a parent of should contain the project's configuration file
            (dg.toml or pyproject.toml with [tool.dg] section).

    Returns:
        Definitions: A merged Definitions object containing all discovered definitions
            from the project's defs folder, enriched with component metadata.

    Example:
        .. code-block:: python

            from pathlib import Path
            import dagster as dg

            @dg.definitions
            def defs():
                project_path = Path("/path/to/my/dg/project")
                return dg.load_from_defs_folder(project_root=project_path)

    """
    path_within_project = normalize_renamed_param(
        path_within_project,
        "path_within_project",
        project_root,
        "project_root",
    )

    return ComponentTree.for_project(
        path_within_project=check.not_none(path_within_project, "Must provide path_within_project")
    ).build_defs()


# Public method so optional Nones are fine
@deprecated(
    breaking_version="1.11",
    additional_warn_text="Use load_from_defs_folder instead.",
)
@suppress_dagster_warnings
def load_defs(
    defs_root: ModuleType,
    project_root: Path | None = None,
    terminate_autoloading_on_keyword_files: bool = True,
) -> Definitions:
    """Constructs a Definitions object, loading all Dagster defs in the given module.

    Args:
        defs_root (Path): The path to the defs root, typically `package.defs`.
        project_root (Optional[Path]): path to the project root directory.
        terminate_autoloading_on_keyword_files (bool): Whether to terminate the defs
            autoloading process when encountering a definitions.py or component.py file.
            Defaults to True.
    """
    project_root = project_root if project_root else get_project_root(defs_root)

    tree = (
        LegacyAutoloadingComponentTree.from_module(defs_module=defs_root, project_root=project_root)
        if terminate_autoloading_on_keyword_files
        else ComponentTree.from_module(defs_module=defs_root, project_root=project_root)
    )

    return tree.build_defs()


def get_library_json_enriched_defs(tree: ComponentTree | None) -> Definitions:
    from dagster.components.core.package_entry import discover_entry_point_package_objects
    from dagster.components.core.snapshot import get_package_entry_snap

    registry_objects = discover_entry_point_package_objects()
    snaps = [get_package_entry_snap(key, obj) for key, obj in registry_objects.items()]
    components_json = json_for_all_components(snaps)

    return Definitions(
        metadata={PLUGIN_COMPONENT_TYPES_JSON_METADATA_KEY: components_json},
        component_tree=tree,
    )



import asyncio
import os

import structlog

from dagster._core.definitions.assets.graph.remote_asset_graph import BaseAssetNode
from dagster._core.definitions.freshness import (
    FreshnessState,
    FreshnessStateChange,
    FreshnessStateRecord,
)
from dagster._core.definitions.freshness_evaluator import FRESHNESS_EVALUATORS_BY_POLICY_TYPE
from dagster._core.workspace.context import IWorkspaceProcessContext, LoadingContext
from dagster._daemon.daemon import IntervalDaemon
from dagster._time import get_current_timestamp

_INTERVAL_SECONDS = int(os.environ.get("FRESHNESS_DAEMON_INTERVAL_SECONDS", "30"))


class FreshnessDaemon(IntervalDaemon):
    def __init__(self, interval_seconds: int = _INTERVAL_SECONDS):
        super().__init__(interval_seconds=interval_seconds)

        self.logger = structlog.wrap_logger(
            self._logger,
        )

    @classmethod
    def daemon_type(cls):
        return "FRESHNESS_DAEMON"

    def run_iteration(self, workspace_process_context: IWorkspaceProcessContext):
        """On every daemon iteration, we try to evaluate freshness for all assets in the asset graph,
        skipping assets that do not have a freshness policy.
        """
        yield
        start_time = get_current_timestamp()
        self.logger.debug("Start freshness daemon iteration", start_timestamp=start_time)

        results = asyncio.run(
            self._evaluate_assets(workspace_process_context=workspace_process_context)
        )
        end_time = get_current_timestamp()

        self.logger.debug(
            f"End freshness daemon iteration.\n Evaluated: {len(results)}\nState Updated: {sum(results)}",
            start_timestamp=start_time,
            end_timestamp=end_time,
            daemon_iteration_seconds=end_time - start_time,
        )

    async def _evaluate_assets(
        self, workspace_process_context: IWorkspaceProcessContext
    ) -> list[bool]:
        """Evaluate the freshness state of all assets in the asset graph.

        Returns:
            A list of booleans indicating whether the freshness state changed for each asset.
        """
        # create a loading context to batch queries against
        loading_context = workspace_process_context.create_request_context()
        asset_graph = loading_context.asset_graph

        coroutines = [
            self._evaluate_asset(context=loading_context, node=node)
            for node in asset_graph.asset_nodes
        ]
        return await asyncio.gather(*coroutines)

    async def _compute_asset_freshness_state(
        self, context: LoadingContext, node: BaseAssetNode
    ) -> FreshnessState:
        """Given an asset spec, return the current freshness state of the asset."""
        # The freshness policy from 1.11 onwards is stored as a first-class property on the asset node.
        # Prior to 1.11, it was stored as a metadata field on the asset spec.
        # We need to support both cases.
        freshness_policy = node.freshness_policy_or_from_metadata
        if not freshness_policy:
            return FreshnessState.NOT_APPLICABLE

        evaluator = FRESHNESS_EVALUATORS_BY_POLICY_TYPE.get(freshness_policy.__class__)
        if not evaluator:
            self.logger.warning(
                f"Asset {node.key} has freshness policy type {freshness_policy.__class__} that has no freshness evaluator"
            )
            return FreshnessState.UNKNOWN

        return await evaluator().evaluate_freshness(context=context, node=node)

    async def _evaluate_asset(self, context: LoadingContext, node: BaseAssetNode) -> bool:
        """Computes the current freshness state of an asset and emits a freshness state change event if the state has changed.

        Returns:
            True if the freshness state has changed, False otherwise.
        """
        previous_state = await FreshnessStateRecord.gen(context, node.key)
        current_state = await self._compute_asset_freshness_state(context=context, node=node)
        if previous_state is None or previous_state.freshness_state != current_state:
            # Note: future refactors could batch the event emission given a new instance method, but freshness state change
            # events are somewhat rare so batching would not improve performance significantly
            context.instance._report_runless_asset_event(  # noqa: SLF001
                asset_event=FreshnessStateChange(
                    key=node.key,
                    previous_state=previous_state.freshness_state
                    if previous_state
                    else FreshnessState.UNKNOWN,
                    new_state=current_state,
                    state_change_timestamp=get_current_timestamp(),
                )
            )
            return True
        return False



from dagster._core.definitions.assets.definition.asset_spec import (
    apply_freshness_policy as apply_freshness_policy,
)
from dagster._core.definitions.freshness import FreshnessPolicy as FreshnessPolicy

__all__ = [
    "FreshnessPolicy",
    "apply_freshness_policy",
]



import os


def get_materialization_chunk_size() -> int:
    """Get the chunk size for fetching materializations after a cursor.

    Returns:
        int: The chunk size for materialization queries, configurable via
        DAGSTER_FETCH_MATERIALIZATIONS_AFTER_CURSOR_CHUNK_SIZE environment variable.
        Defaults to 1000.
    """
    return int(os.getenv("DAGSTER_FETCH_MATERIALIZATIONS_AFTER_CURSOR_CHUNK_SIZE", "1000"))



from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable
from functools import cached_property
from typing import Any, TypeVar

from dagster_shared import check

from dagster._config.field import Field
from dagster._core.definitions.asset_checks.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_checks.asset_checks_definition import AssetChecksDefinition
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.decorators.asset_check_decorator import multi_asset_check
from dagster._core.definitions.decorators.asset_decorator import multi_asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.result import MaterializeResult
from dagster._core.execution.context.asset_check_execution_context import AssetCheckExecutionContext
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.resolved.base import Resolvable
from dagster.components.resolved.core_models import (
    OpSpec,
    ResolvedAssetCheckSpec,
    ResolvedAssetSpec,
)
from dagster.components.resolved.model import Model

T = TypeVar("T", bound=MaterializeResult | AssetCheckResult)


def to_iterable(result: Any, of_type: type[T] | tuple[type[T], ...]) -> Generator[T, None, None]:
    if isinstance(result, of_type):
        yield result
    elif isinstance(result, Iterable):
        for item in result:
            if not isinstance(item, of_type):
                check.failed(f"Expected a {of_type}, got {type(item).__name__}")
            yield item
    elif result:
        check.failed(f"Expected a {of_type}, got {type(result).__name__}")


class ExecutableComponent(Component, Resolvable, Model, ABC):
    """Executable Component represents an executable node in the asset graph.

    It is comprised of an execute_fn, which is can be specified as a fully
    resolved symbol reference in yaml. This makes it a plain ole' Python function
    that does the execution within the asset graph.

    You can pass an arbitrary number of assets or asset checks to the component.

    With this structure this component replaces @asset, @multi_asset, @asset_check, and @multi_asset_check.
    which can all be expressed as a single ExecutableComponent.
    """

    assets: list[ResolvedAssetSpec] | None = None
    checks: list[ResolvedAssetCheckSpec] | None = None

    @property
    @abstractmethod
    def op_spec(self) -> OpSpec: ...

    @property
    def resource_keys(self) -> set[str]:
        return set()

    @cached_property
    def config_fields(self) -> dict[str, Field] | None:
        return None

    @abstractmethod
    def invoke_execute_fn(
        self,
        context: AssetExecutionContext | AssetCheckExecutionContext,
        component_load_context: ComponentLoadContext,
    ) -> Iterable[MaterializeResult | AssetCheckResult]: ...

    def build_underlying_assets_def(
        self, component_load_context: ComponentLoadContext
    ) -> AssetsDefinition:
        if self.assets:

            @multi_asset(
                name=self.op_spec.name,
                op_tags=self.op_spec.tags,
                description=self.op_spec.description,
                specs=self.assets,
                check_specs=self.checks,
                required_resource_keys=self.resource_keys,
                pool=self.op_spec.pool,
                config_schema=self.config_fields,
            )
            def _assets_def(context: AssetExecutionContext, **kwargs):
                return to_iterable(
                    self.invoke_execute_fn(context, component_load_context=component_load_context),
                    of_type=(MaterializeResult, AssetCheckResult),
                )

            return _assets_def
        elif self.checks:

            @multi_asset_check(
                name=self.op_spec.name,
                op_tags=self.op_spec.tags,
                specs=self.checks,
                description=self.op_spec.description,
                required_resource_keys=self.resource_keys,
                pool=self.op_spec.pool,
                config_schema=self.config_fields,
            )
            def _asset_check_def(context: AssetCheckExecutionContext, **kwargs):
                return to_iterable(
                    self.invoke_execute_fn(context, component_load_context=component_load_context),
                    of_type=AssetCheckResult,
                )

            return _asset_check_def

        check.failed("No assets or checks provided")

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        assets_def = self.build_underlying_assets_def(component_load_context=context)
        if isinstance(assets_def, AssetChecksDefinition):
            return Definitions(asset_checks=[assets_def])
        else:
            return Definitions(assets=[assets_def])



"""Scope objects for use in component template resolution."""

import os
import warnings
from typing import Any

from dagster.components.resolved.errors import ResolutionException


class EnvScope:
    """Provides access to environment variables within templates.

    Available via `{{ env.* }}` in component YAML files.
    """

    def __call__(self, key: str, default: str | Any | None = ...) -> str | None:
        value = os.environ.get(key, default=default)
        if value is ...:
            raise ResolutionException(
                f"Environment variable {key} is not set and no default value was provided."
                f" To provide a default value, use e.g. `env('{key}', 'default_value')`."
            )
        return value

    def __getattr__(self, key: str) -> str | None:
        # jinja2 applies a hasattr check to any scope fn - we avoid raising our own exception
        # for this access
        if key.startswith("jinja"):
            raise AttributeError(f"{key} not found")

        return os.environ.get(key)

    def __getitem__(self, key: str) -> str | None:
        raise ResolutionException(
            f"To access environment variables, use dot access or the `env` function, e.g. `env.{key}` or `env('{key}')`"
        )


class WrappedObjectScope:
    """Base class for wrapping an object and exposing its attributes in template scope.

    This allows us to expose Python objects (like modules or context objects) to Jinja templates
    with explicit attribute whitelisting.

    Args:
        wrapped_object: The object to wrap and expose to templates
        accessible_attributes: Set of attribute names that are allowed to be accessed from templates.
                              Only these attributes will be accessible via {{ scope.attr }}
    """

    def __init__(self, wrapped_object: Any, accessible_attributes: set[str]):
        self._wrapped_object = wrapped_object
        self._accessible_attributes = accessible_attributes

    def __getattr__(self, name: str):
        """Allow access to whitelisted wrapped object attributes."""
        # jinja2 applies a hasattr check to any scope fn - we avoid raising our own exception
        # for this access
        if name.startswith("jinja") or name.startswith("_"):
            raise AttributeError(f"{name} not found")
        # Check if this attribute is whitelisted
        if name not in self._accessible_attributes:
            raise AttributeError(
                f"Attribute '{name}' is not accessible. "
                f"Available attributes: {', '.join(sorted(self._accessible_attributes))}"
            )

        return getattr(self._wrapped_object, name)


class DgScope(WrappedObjectScope):
    """Provides access to Dagster definitions and utilities within templates.

    Available via `{{ dg.* }}` in component YAML files.

    Examples:
        {{ dg.AutomationCondition.eager() }}
        {{ dg.DailyPartitionsDefinition(start_date="2024-01-01") }}
    """

    def __init__(self):
        import dagster as dg

        accessible_attributes = {
            "AutomationCondition",
            "FreshnessPolicy",
            "DailyPartitionsDefinition",
            "WeeklyPartitionsDefinition",
            "MonthlyPartitionsDefinition",
            "HourlyPartitionsDefinition",
            "StaticPartitionsDefinition",
            "TimeWindowPartitionsDefinition",
        }

        super().__init__(dg, accessible_attributes)


class DatetimeScope(WrappedObjectScope):
    """Provides access to Python datetime utilities within templates.

    Available via `{{ datetime.* }}` in component YAML files.

    Examples:
        {{ datetime.datetime.now() }}
        {{ datetime.timedelta(days=7) }}
    """

    def __init__(self):
        import datetime

        accessible_attributes = {"datetime", "timedelta"}

        super().__init__(datetime, accessible_attributes)


class LoadContextScope(WrappedObjectScope):
    """Provides access to component loading utilities within templates.

    Available via `{{ context.* }}` in component YAML files.

    Examples:
        {{ context.project_root }}/data/input.csv
        {{ context.load_component("other_component") }}
        {{ context.build_defs("submodule") }}
    """

    def __init__(self, context):
        accessible_attributes = {
            "load_component",
            "build_defs",
            "project_root",
        }

        super().__init__(context, accessible_attributes)

    @property
    def project_root(self) -> str:
        # override here so we can resolve the project_root to a string
        return str(self._wrapped_object.project_root.resolve())


class DeprecatedScope:
    """Wrapper that provides backward compatibility for deprecated scope variables.

    Emits a deprecation warning whenever the variable is accessed.
    """

    def __init__(self, old_name: str, new_name: str, value: Any):
        self._old_name = old_name
        self._new_name = new_name
        self._value = value

    def _warn(self):
        warnings.warn(
            f"Accessing `{{{{{self._old_name}}}}}` is deprecated and will be "
            f"removed in Dagster 1.13.0. Use `{{{{{self._new_name}}}}}` instead.",
            DeprecationWarning,
            stacklevel=4,
        )

    def __call__(self, *args, **kwargs):
        self._warn()
        return self._value(*args, **kwargs)

    def __getattr__(self, key: str):
        # jinja2 applies a hasattr check to any scope fn - we avoid raising our own exception
        # for this access
        if key.startswith("jinja") or key.startswith("_"):
            raise AttributeError(f"{key} not found")

        self._warn()
        return getattr(self._value, key)

    def __getitem__(self, key: str):
        self._warn()
        return self._value[key]

    def __str__(self):
        self._warn()
        return str(self._value)



from dagster._core.storage.defs_state.base import DefsStateStorage as DefsStateStorage
from dagster._core.storage.defs_state.blob_storage_state_storage import (
    UPathDefsStateStorage as UPathDefsStateStorage,
)



"""Settings methods for DagsterInstance."""

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any

import dagster._check as check
from dagster._core.instance.config import (
    DEFAULT_LOCAL_CODE_SERVER_STARTUP_TIMEOUT,
    ConcurrencyConfig,
)

if TYPE_CHECKING:
    from dagster._core.launcher import RunLauncher
    from dagster._core.run_coordinator import RunCoordinator


class SettingsMethods:
    """Mixin class providing settings capabilities for DagsterInstance.

    This class contains all settings-related methods and properties that were
    previously in SettingsMixin, consolidating the functionality without
    the need for separate domain classes.
    """

    # These attributes are provided by DagsterInstance
    _run_monitoring_enabled: bool

    @property
    def is_ephemeral(self) -> bool:
        """Property that should be implemented by the concrete class."""
        raise NotImplementedError

    @property
    def run_launcher(self) -> "RunLauncher":
        """Property that should be implemented by the concrete class."""
        raise NotImplementedError

    @property
    def run_coordinator(self) -> "RunCoordinator":
        """Property that should be implemented by the concrete class."""
        raise NotImplementedError

    def get_settings(self, settings_key: str) -> Any:
        """Abstract method to get settings value - implemented by DagsterInstance."""
        raise NotImplementedError

    def get_backfill_settings(self) -> Mapping[str, Any]:
        return self.get_settings("backfills")

    def get_scheduler_settings(self) -> Mapping[str, Any]:
        return self.get_settings("schedules")

    def get_sensor_settings(self) -> Mapping[str, Any]:
        return self.get_settings("sensors")

    def get_auto_materialize_settings(self) -> Mapping[str, Any]:
        return self.get_settings("auto_materialize")

    @property
    def telemetry_enabled(self) -> bool:
        if self.is_ephemeral:
            return False

        dagster_telemetry_enabled_default = True

        telemetry_settings = self.get_settings("telemetry")

        if not telemetry_settings:
            return dagster_telemetry_enabled_default

        if "enabled" in telemetry_settings:
            return telemetry_settings["enabled"]
        else:
            return dagster_telemetry_enabled_default

    @property
    def nux_enabled(self) -> bool:
        if self.is_ephemeral:
            return False

        nux_enabled_by_default = True

        nux_settings = self.get_settings("nux")
        if not nux_settings:
            return nux_enabled_by_default

        if "enabled" in nux_settings:
            return nux_settings["enabled"]
        else:
            return nux_enabled_by_default

    # run monitoring

    @property
    def run_monitoring_enabled(self) -> bool:
        return self._run_monitoring_enabled

    @property
    def run_monitoring_settings(self) -> Any:
        return self.get_settings("run_monitoring")

    @property
    def run_monitoring_start_timeout_seconds(self) -> int:
        return self.run_monitoring_settings.get("start_timeout_seconds", 180)

    @property
    def run_monitoring_cancel_timeout_seconds(self) -> int:
        return self.run_monitoring_settings.get("cancel_timeout_seconds", 180)

    @property
    def run_monitoring_max_runtime_seconds(self) -> int:
        return self.run_monitoring_settings.get("max_runtime_seconds", 0)

    @property
    def code_server_settings(self) -> Any:
        return self.get_settings("code_servers")

    @property
    def code_server_process_startup_timeout(self) -> int:
        return self.code_server_settings.get(
            "local_startup_timeout", DEFAULT_LOCAL_CODE_SERVER_STARTUP_TIMEOUT
        )

    @property
    def code_server_reload_timeout(self) -> int:
        return self.code_server_settings.get(
            "reload_timeout", DEFAULT_LOCAL_CODE_SERVER_STARTUP_TIMEOUT
        )

    @property
    def wait_for_local_code_server_processes_on_shutdown(self) -> bool:
        return self.code_server_settings.get("wait_for_local_processes_on_shutdown", False)

    @property
    def run_monitoring_max_resume_run_attempts(self) -> int:
        return self.run_monitoring_settings.get("max_resume_run_attempts", 0)

    @property
    def run_monitoring_poll_interval_seconds(self) -> int:
        return self.run_monitoring_settings.get("poll_interval_seconds", 120)

    @property
    def cancellation_thread_poll_interval_seconds(self) -> int:
        return self.get_settings("run_monitoring").get(
            "cancellation_thread_poll_interval_seconds", 10
        )

    @property
    def run_retries_enabled(self) -> bool:
        return self.get_settings("run_retries").get("enabled", False)

    @property
    def run_retries_max_retries(self) -> int:
        return self.get_settings("run_retries").get("max_retries", 0)

    @property
    def run_retries_retry_on_asset_or_op_failure(self) -> bool:
        return self.get_settings("run_retries").get("retry_on_asset_or_op_failure", True)

    @property
    def auto_materialize_enabled(self) -> bool:
        return self.get_settings("auto_materialize").get("enabled", True)

    @property
    def freshness_enabled(self) -> bool:
        return self.get_settings("freshness").get("enabled", True)

    @property
    def auto_materialize_minimum_interval_seconds(self) -> int:
        return self.get_settings("auto_materialize").get("minimum_interval_seconds")

    @property
    def auto_materialize_run_tags(self) -> dict[str, str]:
        return self.get_settings("auto_materialize").get("run_tags", {})

    @property
    def auto_materialize_respect_materialization_data_versions(self) -> bool:
        return self.get_settings("auto_materialize").get(
            "respect_materialization_data_versions", False
        )

    @property
    def auto_materialize_max_tick_retries(self) -> int:
        return self.get_settings("auto_materialize").get("max_tick_retries", 3)

    @property
    def auto_materialize_use_sensors(self) -> bool:
        return self.get_settings("auto_materialize").get("use_sensors", True)

    @property
    def global_op_concurrency_default_limit(self) -> int | None:
        return self.get_concurrency_config().pool_config.default_pool_limit

    # python logs

    @property
    def managed_python_loggers(self) -> Sequence[str]:
        python_log_settings = self.get_settings("python_logs") or {}
        loggers: Sequence[str] = python_log_settings.get("managed_python_loggers", [])
        return loggers

    @property
    def python_log_level(self) -> str | None:
        python_log_settings = self.get_settings("python_logs") or {}
        return python_log_settings.get("python_log_level")

    def _initialize_run_monitoring(self) -> None:
        """Initialize run monitoring settings and validate configuration."""
        run_monitoring_enabled = self.run_monitoring_settings.get("enabled", False)
        self._run_monitoring_enabled = run_monitoring_enabled
        if self.run_monitoring_enabled and self.run_monitoring_max_resume_run_attempts:
            check.invariant(
                self.run_launcher.supports_resume_run,
                "The configured run launcher does not support resuming runs. Set"
                " max_resume_run_attempts to 0 to use run monitoring. Any runs with a failed"
                " run worker will be marked as failed, but will not be resumed.",
            )

    def get_python_log_dagster_handler_config(self) -> dict:
        """Extract dagster handler config from python_logs settings."""
        return self.get_settings("python_logs").get("dagster_handler_config", {})

    def get_concurrency_config(self) -> ConcurrencyConfig:
        """Get concurrency configuration from settings."""
        from dagster._core.run_coordinator.queued_run_coordinator import QueuedRunCoordinator

        if isinstance(self.run_coordinator, QueuedRunCoordinator):
            run_coordinator_run_queue_config = self.run_coordinator.get_run_queue_config()
        else:
            run_coordinator_run_queue_config = None

        concurrency_settings = self.get_settings("concurrency")
        return ConcurrencyConfig.from_concurrency_settings(
            concurrency_settings, run_coordinator_run_queue_config
        )



import importlib.util
import subprocess
import sys
import textwrap
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from types import ModuleType
from typing import TypeVar

import click
from dagster_shared.error import DagsterError

from dagster import _check as check
from dagster.components.utils.translation import TranslatorResolvingInfo as TranslatorResolvingInfo

T = TypeVar("T")


def exit_with_error(error_message: str) -> None:
    click.echo(click.style(error_message, fg="red"))
    sys.exit(1)


def format_error_message(message: str) -> str:
    # width=10000 unwraps any hardwrapping
    dedented = textwrap.dedent(message).strip()
    paragraphs = [textwrap.fill(p, width=10000) for p in dedented.split("\n\n")]
    return "\n\n".join(paragraphs)


# Temporarily places a path at the front of sys.path, ensuring that any modules in that path are
# importable.
@contextmanager
def ensure_loadable_path(path: Path) -> Iterator[None]:
    orig_path = sys.path.copy()
    sys.path.insert(0, str(path))
    try:
        yield
    finally:
        sys.path = orig_path


def get_path_for_package(package_name: str) -> str:
    spec = importlib.util.find_spec(package_name)
    if not spec:
        raise DagsterError(f"Cannot find package: {package_name}")
    # file_path = spec.origin
    submodule_search_locations = spec.submodule_search_locations
    if not submodule_search_locations:
        raise DagsterError(f"Package does not have any locations for submodules: {package_name}")
    return submodule_search_locations[0]


# ########################
# ##### PLATFORM
# ########################


def is_windows() -> bool:
    return sys.platform == "win32"


def is_macos() -> bool:
    return sys.platform == "darwin"


# ########################
# ##### VENV
# ########################


def get_venv_executable(venv_dir: Path, executable: str = "python") -> Path:
    if is_windows():
        return venv_dir / "Scripts" / f"{executable}.exe"
    else:
        return venv_dir / "bin" / executable


def install_to_venv(venv_dir: Path, install_args: list[str]) -> None:
    executable = get_venv_executable(venv_dir)
    command = ["uv", "pip", "install", "--python", str(executable), *install_args]
    subprocess.run(command, check=True)


def get_path_from_module(module: ModuleType) -> Path:
    module_path = (
        Path(module.__file__).parent
        if module.__file__
        else Path(module.__path__[0])
        if module.__path__
        else None
    )
    return check.not_none(module_path, f"Module {module.__name__} has no filepath")
